"""
Provides elements that allow participant input.

.. moduleauthor: Johannes Brachem <jbrachem@posteo.de>
"""

import re
from datetime import datetime
from typing import List, Tuple, Union

import bleach
import cmarkgfm
from cmarkgfm.cmark import Options as cmarkgfmOptions
from emoji import emojize

from .._helper import inherit_kwargs
from ..exceptions import AlfredError
from .core import ChoiceElement, Element, InputElement, _Choice, jinja_env


@inherit_kwargs
class TextEntry(InputElement):
    """
    Provides a text entry field.

    Args:
        placeholder: Placeholder text, displayed inside the input field.
        {kwargs}

    Examples:
        ::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.TextEntry(toplab="Enter here", name="el1")

    """

    element_template = jinja_env.get_template("html/TextEntryElement.html.j2")

    def __init__(
        self,
        placeholder: str = None,
        **kwargs,
    ):
        """Constructor method."""
        super().__init__(**kwargs)

        self._placeholder = placeholder if placeholder is not None else ""

    @property
    def input(self) -> str:
        # docstring inherited
        return self._input

    @input.setter
    def input(self, value):
        if not value:
            self._input = None
        else:
            self._input = bleach.clean(value)  # sanitizing input

    @property
    def placeholder(self) -> str:
        """str: Placeholder text, displayed inside the input field."""
        return self._placeholder

    @placeholder.setter
    def placeholder(self, value):
        self._placeholder = value

    @property
    def template_data(self):

        d = super().template_data
        d["placeholder"] = self.placeholder
        return d

    @property
    def codebook_data(self):

        data = super().codebook_data
        data["placeholder"] = self.placeholder
        return data


@inherit_kwargs
class TextArea(TextEntry):
    """
    A text area for long text input.

    Args:
        nrows: Initial height of the text area in number of rows.
        {kwargs}

    Examples:
        ::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.TextArea(toplab="Enter here", name="el1")

    """

    element_template = jinja_env.get_template("html/TextAreaElement.html.j2")

    def __init__(self, nrows: int = 5, **kwargs):
        super().__init__(**kwargs)
        self.area_nrows = nrows

    @property
    def template_data(self):

        d = super().template_data
        d["area_nrows"] = self.area_nrows
        return d


@inherit_kwargs
class MatchEntry(TextEntry):
    """
    Displays an input field, which only accepts inputs, matching a
    predefined regular expression.

    Args:
        pattern: Regular expression string to match with user input. We
            recommend that you use raw literal strings prefaced by 'r'
            (see Examples). That removes the necessity for escaping some
            characters twice and thus makes the regular expression
            easier to read and write.
        match_hint: Hint to be displayed if the user input doesn't
            match with the regular expression.
        {kwargs}

    .. versionadded:: 2.3.0

    Examples:

        Example for a MatchEntry element that will match any input::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.MatchEntry(toplab="Enter text here", pattern=r".*", name="reg1")

    """

    def __init__(self, pattern: str = r".*", match_hint: str = None, **kwargs):
        super().__init__(**kwargs)

        #: Compiled regular expression pattern to be matched on
        #: participant input to this element
        self.pattern: re.Pattern = re.compile(pattern)
        self._match_hint = match_hint  # documented in getter property

    def validate_data(self, silent: bool = False) -> bool:

        if not self.should_be_shown:
            return True

        elif not self.force_input and not self.input:
            return True

        elif not self.input:
            if not silent:
                self.hint_manager.post_message(self.no_input_hint)
            return False

        elif not self.pattern.fullmatch(self.input):
            if not silent:
                self.hint_manager.post_message(self.match_hint)
            return False

        else:
            return True

    @property
    def match_hint(self):
        """
        str: Hint to be displayed, if participant input does not match
        the provided pattern.
        """
        if self._match_hint:
            return self._match_hint
        else:
            return self.default_match_hint

    @property
    def default_match_hint(self) -> str:
        """
        str: Default match hint for this element, extracted from config.conf
        """
        name = f"match_{type(self).__name__}"
        return self.experiment.config.get("hints", name)

    @property
    def codebook_data(self) -> dict:

        d = super().codebook_data
        d["regex_pattern"] = self.pattern.pattern
        return d


@inherit_kwargs
class RegEntry(MatchEntry):
    """
    Displays an input field, which only accepts inputs, matching a
    predefined regular expression.

    Args:
        {kwargs}

    Notes:
        Alias for :class:`.MatchEntry`

    Examples:

        Example for a RegEntry element that will match any input::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.RegEntry(toplab="Enter text here", pattern=r".*", name="reg1")

    """

    pass


@inherit_kwargs
class EmailEntry(MatchEntry):
    """
    An input field for email adresses.

    Args:
        {kwargs}

    .. versionadded:: 2.3.0

    Notes:
        This element is a direct descendant of :class:`.MatchEntry`,
        simply adding a default regular expression for matching email
        adresses. Note that there may be different preferences about
        which expression to use and feel free to use your own.

        The one we use here originates from https://emailregex.com


    Examples:
        Example for an EmailEntry element::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.EmailEntry(toplab="Enter text here", name="email")

    """

    def __init__(
        self, pattern=r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)", **kwargs
    ):
        super().__init__(pattern=pattern, **kwargs)


@inherit_kwargs(exclude=["force_input", "pattern"])
class PasswordEntry(RegEntry):
    """
    Displays a password field.

    The password field is **force-input by default**.

    Args:
        password (str): Password string to match against user input.

        force_input (bool): If `True`, users can  only progress to the next page
            if they enter data into this field. Note that a
            :class:`.NoValidationSection` or similar sections might
            overrule this setting. Defaults to *True*.

        {kwargs}


    .. warning:: Note that the password will be included in the
        automatically generated codebook.

    """

    element_template = jinja_env.get_template("html/PasswordEntry.html.j2")

    def __init__(
        self, password: str, force_input: bool = True, match_hint: str = None, **kwargs
    ):
        super(RegEntry, self).__init__(force_input=force_input, **kwargs)
        self.password = password
        self._match_hint = match_hint  # documented in getter property

        if not isinstance(password, str):
            raise ValueError(
                f"Argument 'password' in {type(self).__name__} element '{self.name}'"
                " must be a string."
            )

    def validate_data(self, silent: bool = False) -> bool:

        if not self.should_be_shown:
            return True

        elif not self.force_input and self.input == "":
            return True

        elif not self.input:
            if not silent:
                self.hint_manager.post_message(self.no_input_hint)
            return False

        elif not self.input == self.password:
            if not silent:
                self.hint_manager.post_message(self.match_hint)
            return False

        else:
            return True

    @property
    def codebook_data(self) -> dict:
        d = super(RegEntry, self).codebook_data
        d["password"] = self.password
        return d


@inherit_kwargs(exclude=["force_input", "pattern"])
class MultiplePasswordEntry(RegEntry):
    """
    Password field that accepts multiple different passwords.

    The password field is force-entry by default.

    Args:
        passwords (list): List of password strings to match against user
            input.

        force_input (bool): If `True`, users can  only progress to the next page
            if they enter data into this field. Note that a
            :class:`.NoValidationSection` or similar sections might
            overrule this setting. Defaults to *True*.

        {kwargs}


    .. warning:: Note that the password will be included in the
        automatically generated codebook.

    """

    element_template = jinja_env.get_template("html/PasswordEntry.html.j2")

    def __init__(
        self,
        passwords: List[str],
        force_input: bool = True,
        match_hint: str = None,
        **kwargs,
    ):
        super(RegEntry, self).__init__(force_input=force_input, **kwargs)
        self.passwords = passwords
        self._match_hint = match_hint  # documented in getter property

        if not isinstance(passwords, (list, tuple)):
            raise ValueError(
                f"Argument 'passwords' in {type(self).__name__} element '{self.name}'"
                " must be a list or a tuple."
            )

        for pw in passwords:
            if not isinstance(pw, str):
                raise ValueError(
                    "All elements of the sequence 'passwords' in"
                    f" {type(self).__name__} must be strings."
                )

    def validate_data(self, silent: bool = False) -> bool:

        if not self.should_be_shown:
            return True

        elif not self.force_input and self.input == "":
            return True

        elif not self.input:
            if not silent:
                self.hint_manager.post_message(self.no_input_hint)
            return False

        elif self.input not in self.passwords:
            self.hint_manager.post_message(self.match_hint)
            return False

        else:
            return True

    @property
    def codebook_data(self) -> dict:
        d = super(RegEntry, self).codebook_data
        d["passwords"] = "; ".join(self.passwords)
        return d


@inherit_kwargs
class NumberEntry(TextEntry):
    """
    Displays an input field which only accepts numerical input.

    Args:
        ndecimals (int): Accepted number of decimals. Defaults to 0.
        min (int, float): Minimum accepted entry value. If *None*
            (default), no minimum value is enforced.
        max (int, float): Maximum accepted entry value. If *None*
            (default), no maximum value is enforced.
        decimal_signs (tuple): Tuple of accepted decimal signs. Defaults to
            ``(",", ".")``, i.e. by default, **both** a comma and a dot are
            interpreted as decimal signs. This is due to alfred3 being
            developed in Germany, which uses the comma as a decimal sign.
        match_hint (str): Specialized match hint for this element. You can
            use the placeholders ``{{min}}``, ``{{max}}``, ``{{ndecimals}}``,
            and ``{{decimal_signs}}``. To customize the match hint for
            all NumberEntry elements, change the respective setting
            in the config.conf.
        {kwargs}

    .. versionchanged:: 2.3.0
        NumberEntry element now returns its input as numeric (float).

    Examples:

        Using a NumberEntry element::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                def on_exp_access(self):
                    self += al.NumberEntry(placeholder="Enter a number", name="num1")

    """

    def __init__(
        self,
        ndecimals: int = 0,
        min: Union[int, float] = None,
        max: Union[int, float] = None,
        decimal_signs: Union[str, tuple] = (",", "."),
        match_hint: str = None,
        **kwargs,
    ):

        self.ndecimals: int = ndecimals  # documented in getter property
        self.decimal_signs: Tuple[str] = decimal_signs  # documented in getter property
        self.min = min  # documented in getter property
        self.max = max  # documented in getter property
        self._match_hint = match_hint  # documented in getter property
        self._original_input = None
        super().__init__(**kwargs)

    @property
    def ndecimals(self) -> int:
        """int: Number of allowed decimal places."""
        return self._ndecimals

    @ndecimals.setter
    def ndecimals(self, value: int):
        if not isinstance(value, int) or not value >= 0:
            raise ValueError("Number of decimals must be an integer >= 0.")
        else:
            self._ndecimals = value

    @property
    def decimal_signs(self) -> Union[str, tuple]:
        """Union[str, tuple]: Interpreted decimal signs."""
        return self._decimal_signs

    @decimal_signs.setter
    def decimal_signs(self, value: Union[str, tuple]):
        msg = "Decimals signs must be a string or a tuple of strings."
        if not isinstance(value, (str, tuple)):
            raise ValueError(msg)

        if isinstance(value, tuple):
            for val in value:
                if not isinstance(val, str):
                    raise ValueError(msg)

        self._decimal_signs = value

    @property
    def min(self) -> Union[int, float]:
        """Union[int, float]: Minimum value that is accepted by this element."""
        return self._min

    @min.setter
    def min(self, value: Union[int, float]):
        if value is None:
            self._min = None
        elif not isinstance(value, (int, float)):
            raise ValueError("Minimum must be a number.")
        else:
            self._min = value

    @property
    def max(self) -> Union[int, float]:
        """Union[int, float]: Maximum value that is accepted by this element."""
        return self._max

    @max.setter
    def max(self, value: Union[int, float]):
        if value is None:
            self._max = None
        elif not isinstance(value, (int, float)):
            raise ValueError("Maximum must be a number.")
        else:
            self._max = value

    @property
    def match_hint(self):
        """
        str: Hint to be displayed, if participant input does not match
        the provided pattern.
        """
        if self._match_hint:
            hint = self._match_hint
        else:
            hint = self.default_match_hint

        signs = " ".join([f"<code>{sign}</code>" for sign in self.decimal_signs])

        hint = hint.format(
            min=str(self.min),
            max=str(self.max),
            decimal_signs=signs,
            ndecimals=str(self.ndecimals),
        )

        return hint

    @property
    def default_match_hint(self) -> str:
        """
        str: Default match hint for this element, extracted from config.conf

        This property combines all match hints related to the
        NumberEntry element and returns them as a single string.
        """
        name = f"{type(self).__name__}"

        c = self.experiment.config
        hints = []
        hints.append(c.get("hints", "match_" + name))

        if self.min is not None:
            hints.append(c.get("hints", "min_" + name))

        if self.max is not None:
            hints.append(c.get("hints", "max_" + name))

        hints.append(c.get("hints", "ndecimals_" + name))

        if self.ndecimals > 0 and self.decimal_signs is not None:
            hints.append(c.get("hints", "decimal_signs_" + name))

        return " ".join(hints)

    @property
    def input(self) -> float:
        # docstring inherited
        try:
            return float(self._input) if self._input is not None else None
        except ValueError:
            return self._input

    @input.setter
    def input(self, value: str):
        self._original_input = value
        if not value:
            self._input = None
        else:
            value = str(value)
            for sign in self.decimal_signs:
                value = value.replace(sign, ".")
            self._input = value

    def validate_data(self, silent: bool = False) -> bool:

        if not self.should_be_shown:
            return True

        if not self.force_input and not self._input:
            return True

        elif not self._input:
            if not silent:
                self.hint_manager.post_message(self.no_input_hint)
            return False

        try:
            in_number = float(self._input)
        except ValueError:
            self.hint_manager.post_message(self.match_hint)
            return False

        validate = True
        decimals = self._input.split(".")[-1] if "." in self._input else ""
        if self.min is not None and in_number < self.min:
            self.hint_manager.post_message(self.match_hint)
            validate = False

        elif self.max is not None and in_number > self.max:
            self.hint_manager.post_message(self.match_hint)
            validate = False

        elif len(decimals) > self.ndecimals:
            self.hint_manager.post_message(self.match_hint)
            validate = False
        return validate

    @property
    def codebook_data(self):

        data = super().codebook_data
        data["value"] = self.input

        data["ndecimals"] = self.ndecimals
        data["decimal_signs"] = " ".join(self.decimal_signs)
        data["min"] = self.min
        data["max"] = self.max

        return data

    @property
    def template_data(self):
        d = super().template_data
        d["input"] = self._original_input
        return d


@inherit_kwargs(exclude=["height"])
class RangeInput(InputElement):
    """
    Range input slider for numerical input.

    Args:
        min (float, int): The value won't be less than min. The default
            is 0.
        max (float, int): The value won't be greater than max. The default
            is 100.
        step (float, int): The value will be a multiple of step. The
            default is 1.
        display_input (bool): If *True*, the current input value will be
            displayed alongside the element.
        display_position (str): Position of the input value display.
            Can be *top* or *bottom*. Defaults to *top*.
        align (str): Horizontal alignment of input value display. Can
            be *left*, *right*, and *center*. Defaults to *center*.
        font_size (str, int): Font size for value display. You can use a
            keyword or an exact specification. The available keywords
            are *tiny*, *small*, *normal*, *big*, and *huge*. The exact
            specification shoul ideally include a unit, such as *1rem*,
            or *12pt*. If you supply an integer without a unit, a unit
            of *pt* will be assumed. Defaults to *normal*.
        display_locale (str): A locale specification for
            displaying the current input value in an appropriate format.
            The default is *en-GB*, which uses a dot as a decimal sign.
            Use *de-DE* for german display with a comma as the decimal
            sign. Other possible values can be taken from
            https://developer.mozilla.org/de/docs/Web/JavaScript/Reference/Global_Objects/Intl#locale_identification_and_negotiation
        display_suffix (str): A suffix for the display of the
            current input value. Can be used, for example, to add a
            unit to the display. Defaults to an empty string.
        mindecimals (int): Minimum number of decimals to display.
            Defaults to 0.
        maxdecimals (int): Maximum number of decimals to display.
            Defaults to 2.
        {kwargs}

    Examples:
        Basic example::

            import alfred3 as al

            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.RangeInput(name="range_demo")
    """

    element_template = jinja_env.get_template("html/RangeInputElement.html.j2")
    js_template = jinja_env.get_template("js/range_input.js.j2")

    def __init__(
        self,
        min: Union[float, int] = 0,
        max: Union[float, int] = 100,
        step: Union[float, int] = 1,
        display_input: bool = True,
        display_position: str = "top",
        align: str = "center",
        display_locale: str = "en-GB",
        display_suffix: str = "",
        mindecimals: int = 0,
        maxdecimals: int = 2,
        **kwargs,
    ):
        super().__init__(align=align, **kwargs)
        if display_position not in ["top", "bottom"]:
            raise ValueError(
                f"{self} accepts only 'top' and 'bottom' for argument"
                f" 'display_position'. You gave '{display_position}'."
            )

        self.min = min
        self.max = max
        self.display_position = display_position
        self.display_locale = display_locale
        self.display_suffix = display_suffix
        self.step = step
        self.mindecimals = mindecimals
        self.maxdecimals = maxdecimals
        self.display_input = display_input
        self.offset_display_height = (
            "true" if self.leftlab or self.rightlab else "false"
        )  # for javascript

    def added_to_page(self, page):
        super().added_to_page(page)
        if self.display_input:
            js = self.js_template.render(self.js_template_data)
            self.add_js(js)

    @property
    def js_template_data(self):
        d = {}
        d["name"] = self.name
        d["display_position"] = self.display_position
        d["display_locale"] = self.display_locale
        d["display_suffix"] = self.display_suffix
        d["offset_display_height"] = self.offset_display_height
        d["mindecimals"] = self.mindecimals
        d["maxdecimals"] = self.maxdecimals
        return d

    @property
    def template_data(self):
        d = super().template_data
        d["min"] = self.min
        d["max"] = self.max
        d["step"] = self.step
        d["display_position"] = self.display_position
        return d

    @property
    def codebook_data(self):

        data = super().codebook_data
        data["value"] = self.input
        data["step"] = self.step
        data["display_input"] = self.display_input
        data["min"] = self.min
        data["max"] = self.max

        return data

    @property
    def no_input_hint(self) -> str:
        """
        RangeInput always has an input.
        """
        return "No Input to RangeInput"

    @property
    def match_hint(self):
        """
        Should never be needed for RangeInput
        """
        return "No match for RangeInput"

    @property
    def input(self) -> float:
        # docstring inherited
        return float(self._input) if self._input is not None else None

    @input.setter
    def input(self, value: str):
        if not value:
            self._input = None
        else:
            self._input = value

    def validate_data(self, silent: bool = False) -> bool:

        if not self.should_be_shown:
            return True

        if not self.force_input and not self._input:
            return True

        elif not self.input:
            if not silent:
                self.hint_manager.post_message(self.no_input_hint)
            return False

        try:
            in_number = float(self._input)
        except ValueError:
            self.hint_manager.post_message(self.match_hint)
            return False

        validate = True

        if self.min is not None and in_number < self.min:
            self.hint_manager.post_message(self.match_hint)
            validate = False

        elif self.max is not None and in_number > self.max:
            self.hint_manager.post_message(self.match_hint)
            validate = False

        return validate


@inherit_kwargs
class SingleChoice(ChoiceElement):
    """
    Radio buttons for choosing a single option.

    Args:
        *choice_labels: Variable numbers of choice labels. See
            :class:`.ChoiceElement` for details.
        default: The *default* argument of single choice elements is an
            integer, indicating which choice should be selected by
            default. Counting of choices starts at 1.
        {kwargs}

    Notes:
        This element saves answers in the form of integers, counting
        up from 1. Take a look at the examples to see how to work with
        user input to a SingleChoice element.

    See Also:

        - :class:`.SingleChoiceButtons` and :class:`.SingleChoiceBar` are
          nicer displays of SingleChoice elements. Those are also
          more suitable for responsive displays.
        - The class :class:`.SubmittingButtons` is a version of
          SingleChoiceButtons that automatically moves the experiment
          to the next page, once the participant clicks on an answer.
        - :class:`.SingleChoiceList` is a dropdown list of single
          choices. The SingleChoiceList saves its data as strings
          (instead of integers) and is a good option if you have ten
          or more choices.
        - :class:`.MultipleChoiceButtons` can be used to allow multiple
          answers.


    Examples:
        A simple SingleChoice element::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo_page"

                def on_exp_access(self):
                    self += al.SingleChoice("Yes", "No", name="c1")

        Accessing the input to a SingleChoice element::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo_page"

                def on_exp_access(self):
                    self += al.SingleChoice("Yes", "No", toplab="Choose one", name="c1")


            @exp.member
            class Show(al.Page):

                def on_first_show(self):
                    c1_answer = self.exp.values["c1"] # access value
                    self += al.Text(f"Your answer was: {{c1_answer}}")


    """

    # Documented at :class:`.ChoiceElement`
    type: str = "radio"

    def define_choices(self) -> List[_Choice]:

        choices = []
        for i, label in enumerate(self.choice_labels, start=1):
            choice = _Choice()

            if isinstance(label, Element):
                choice.label = label.web_widget
            else:
                if self.emojize:
                    label = emojize(str(label), use_aliases=True)
                choice.label = cmarkgfm.github_flavored_markdown_to_html(
                    str(label), options=cmarkgfmOptions.CMARK_OPT_UNSAFE
                )
            choice.type = "radio"
            choice.value = i
            choice.name = self.name
            choice.id = f"{self.name}_choice{i}"
            if choice.id in self.exp.root_section.all_elements:
                msg = (
                    f"You have a SingleChoice-type element of name {self.name}, which"
                    f" means that the name '{choice.id}' must be reserved. Please check"
                    " if you are using it for any other element."
                )
                raise AlfredError(msg)

            choice.label_id = f"{choice.id}-lab"
            choice.disabled = True if self.disabled else False

            if self.input:
                choice.checked = i == self.input
            elif self.default is not None:
                choice.checked = True if self.default == str(i) else False

            choice.css_class = f"choice-button choice-button-{self.name}"

            choices.append(choice)
        return choices

    @property
    def default_no_input_hint(self) -> str:
        # docstring inherited
        return self.experiment.config.get("hints", "no_inputSingleChoice")

    def set_data(self, d):

        # Important: We need to have a check that ensures that the
        # generated name for the value of each choice is not used by
        # any other name in the experiment.
        # For this element, we implement it in the *define_choices*
        # method.
        if self.name not in d:
            return
        else:
            self.input = d[self.name]

    @property
    def input(self) -> int:
        """
        int: Index of selected choice (starting at 1). Returns *None*,
        if there is no input.
        """
        return self._input

    @input.setter
    def input(self, value):
        if not value:
            self._input = None
        else:
            self._input = int(value)


@inherit_kwargs
class MultipleChoice(ChoiceElement):
    """
    Checkboxes for choosing multiple options.

    Args:
        *choice_labels: Variable numbers of choice labels. See
            :class:`.ChoiceElement` for details.
        min, max: Minimum (maximum) number of choices that need to be
            selected, if any are selected (does not imply
            *force_input=True*).
        select_hint: Hint to be displayed, if the requirement of
            minimum or maximum number of selected fields is not reached.
            Defaults to the experiment-wide value specified in
            config.conf.
        default: Can be a single integer, or a list of integers which
            indicate the choices that should be selected by default.
            Counting starts at 1.
        {kwargs}

    Notes:
        This element saves and returns not a single value, but a
        dictionary of values. Each choice is represented by a key, and
        the corresponding value is *True*, if the choice was selected and
        *False* otherwise.

        The keys are of the form ``choicei``, where ``i`` is a placeholer
        for the number of the choice, i.e. ``choice1`` for the first choice.

    See Also:
        - See :class:`.MultipleChoiceButtons` and :class:`.MultipleChoiceBar`
          for nice-looking buttons and a button-bar of multiple choices.
        - See :class:`.SingleChoice`, :class:`.SingleChoiceButtons`, and
          :class:`.SingleChoiceBar` for single choice elements.
        - See :class:`.SubmittingButtons` for single choice buttons
          that trigger a forward move on click.

    Examples:
        A multiple choice element with three options::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo_page"

                def on_exp_access(self):
                    self += al.MultipleChoice("Yes", "No", "Maybe", name="m1")

        Accessing the input to a MultipleChoice element::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.MultipleChoice("a", "b", "c", toplab="Choose one or more", name="c1")

            @exp.member
            class Show(al.Page):

                def on_first_show(self):
                    choices = self.exp.values.get("c1")
                    self += al.Text(f"Your answer is saved like this: {{choices}}")

    """

    type = "checkbox"

    def __init__(
        self,
        *choice_labels,
        min: int = None,
        max: int = None,
        select_hint: str = None,
        default: Union[int, List[int]] = None,
        **kwargs,
    ):
        super().__init__(*choice_labels, **kwargs)

        self._input = {}

        #: See documentation of initialization arguments
        self.min: int = min if min is not None else 0

        #: See documentation of initialization arguments
        self.max: int = max if max is not None else len(self.choice_labels)

        self._select_hint = select_hint  # documented in getter method

        if isinstance(default, int):
            self.default = [default]
        elif default is not None and not isinstance(default, list):
            raise ValueError(
                "Default for MultipleChoice must be a list of integers, indicating the"
                " default choices."
            )
        else:
            self.default = default

    @property
    def select_hint(self) -> str:
        """
        str: Hint for participants. Displayed, if the number of selected
        choices does not fulfill the requirements of the *min* and *max*
        number.
        """
        if self._select_hint:
            return self._select_hint
        else:
            msg = self.experiment.config.get("hints", "select_MultipleChoice")
            return msg.format(min=self.min, max=self.max)

    def validate_data(self, silent: bool = False) -> bool:
        checked_values = {k: v for k, v in self.input.items() if v}

        if not self.should_be_shown:
            return True

        elif self.force_input and not checked_values:
            self.hint_manager.post_message(self.no_input_hint)
            return False

        elif not (self.min <= sum(list(self.input.values())) <= self.max):
            self.hint_manager.post_message(self.select_hint)
            return False

        else:
            return True

    def set_data(self, d):

        # Important: We need to have a check that ensures that the
        # generated name for the value of each choice is not used by
        # any other name in the experiment.
        # For this element, we implement it in the *define_choices*
        # method.
        for choice in self.choices:

            value = d.get(choice.name, None)
            self._input[f"choice{choice.value}"] = str(choice.value) == value

    def define_choices(self):

        choices = []
        for i, label in enumerate(self.choice_labels, start=1):
            choice = _Choice()

            if isinstance(label, Element):
                choice.label = label.web_widget
            else:
                if self.emojize:
                    label = emojize(str(label), use_aliases=True)
                choice.label = cmarkgfm.github_flavored_markdown_to_html(
                    str(label), options=cmarkgfmOptions.CMARK_OPT_UNSAFE
                )
            choice.type = "checkbox"
            choice.value = i
            choice.id = f"{self.name}_choice{i}"

            if choice.id in self.exp.root_section.all_elements:
                msg = (
                    f"You have a MultipleChoice-type element of name {self.name}, which"
                    f" means that the name '{choice.id}' must be reserved. Please check"
                    " if you are using it for any other element."
                )
                raise AlfredError(msg)

            choice.name = choice.id
            choice.label_id = f"{choice.id}-lab"
            choice.css_class = f"choice-button choice-button-{self.name}"

            if self.debug_enabled:
                choice.checked = True if i <= self.max else False
            elif self.input:
                choice.checked = self.input[f"choice{i}"]
            elif self.default:
                choice.checked = True if i in self.default else False

            choices.append(choice)
        return choices

    @property
    def default_no_input_hint(self) -> str:
        # docstring inherited
        return self.experiment.config.get("hints", "no_inputMultipleChoice")

    @property
    def input(self) -> dict:
        """
        Dict[str, bool]: Dictionary of subject inputs. Returns an empty
        dictionary, if there is no input.

        The keys are 'choice{{i}}', with {{i}} being replaced by the
        choice number. The values are *True* or *False*, indicating
        whether the respective choice has been selected.

        """
        return self._input

    @input.setter
    def input(self, value):
        self._input = value


@inherit_kwargs(exclude=["align"])
class SingleChoiceList(SingleChoice):
    """
    A dropdown list, allowing selection of one option.

    Args:
        *choice_labels: Variable numbers of choice labels. Labels must be
            strings. The number of labels determines the number of choices.
        {kwargs}

    Notes:
        The SingleChoiceList's input always defaults to the first option.
        A typical way to remove meaning from this default is
        to make the fist choice a no-choice option (see examples).

        Different from pure :class:`.SingleChoice` elements, the
        SingleChoiceList saves its input as strings. The examples below
        illustrate how to work with user input.

    See Also:

        - :class:`.SingleChoiceButtons`, :class:`.SingleChoiceBar`,
          and class:`.SingleChoice` for alternatives for selecting
          a single option.
        - The class :class:`.SubmittingButtons` is a version of
          SingleChoiceButtons that automatically moves the experiment
          to the next page, once the participant clicks on an answer.
        - :class:`.MultipleChoiceButtons` can be used to allow multiple
          answers.

    Examples:
        A single choice list with a no-choice option as first option::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.SingleChoiceList(
                        "-no selection-", "choi1", "choi2", "choi3",
                         name="sel1"
                         )


        A single choice list with a no-choice option as first option that
        still enforces force-input. This works, because the empty string
        will not be accepted if force_input is True::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.SingleChoiceList(
                        "", "choi1", "choi2", "choi3",
                         name="sel1", force_input=True
                         )


        Accessing the value of a SingleChoiceList::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.SingleChoiceList(
                        "-no selection-", "choi1", "choi2", "choi3",
                         name="sel1"
                         )

            @exp.member
            class Show(al.Page):

                def on_first_show(self):
                    selection = self.exp.values["sel1"] # accesses selection
                    self += al.Text(f"You selected: {{selection}}")

    """

    # Documented at :class:`.Element`
    element_template = jinja_env.get_template("html/SelectElement.html.j2")

    # Documented at :class:`.SingleChoice`
    type = "select_one"

    def __init__(self, *choice_labels, default: int = 1, **kwargs):
        if "align" in kwargs:
            raise ValueError(
                f"Argument 'align' is not supported for {type(self).__name__}"
            )
        super().__init__(*choice_labels, default=default, **kwargs)

    def define_choices(self) -> List[_Choice]:

        choices = []
        for i, label in enumerate(self.choice_labels, start=1):
            choice = _Choice()

            if not isinstance(label, str):
                raise TypeError(
                    f"Choice label in {type(self).__name__} must be string, not"
                    f" {type(label)}."
                )

            choice.label = label
            choice.type = "radio"
            choice.value = choice.label
            choice.name = self.name
            choice.id = f"{self.name}_choice{i}"
            if choice.id in self.exp.root_section.all_elements:
                msg = (
                    f"You have a SingleChoice-type element of name {self.name}, which"
                    f" means that the name '{choice.id}' must be reserved. Please check"
                    " if you are using it for any other element."
                )
                raise AlfredError(msg)

            choice.label_id = f"{choice.id}-lab"
            choice.disabled = True if self.disabled else False

            # set default
            if isinstance(self.input, int):
                choice.checked = self.input == i
            else:
                choice.checked = self.input == choice.value

            choice.css_class = f"choice-button choice-button-{self.name}"

            choices.append(choice)
        return choices

    @property
    def input(self) -> str:
        """
        str: Text of selected choice. Returns *None*,
        if there is no input.

        Note that, differing from :class:`.SingleChoice`, we do not
        use an index here. This is due to the fact that in
        :class:`.SingleChoiceList`, all labels must be strings, while
        they can have other classes in SingleChoice elements. Thus, it
        is safe to use the label in SingleChoiceLists, but not in
        ordinary SingleChoice elements.
        """
        return self._input

    @input.setter
    def input(self, value):
        if not value:
            self._input = None
        else:
            self._input = value


# @inherit_kwargs
# class MultipleChoiceList(MultipleChoice):
#     """
#     A :class:`.MultipleChoice` element, displayed as list.

#     Args:
#         *choice_labels: Variable numbers of choice labels. See
#             :class:`.ChoiceElement` for details.
#         size: The vertical height of the list. The unit is the number
#             of choices to be displayed without scrolling. Note that some
#             browsers do not adhere to this unit exactly.
#         {kwargs}

#     Examples:
#         Minimal example::

#             import alfred3 as al
#             exp = al.Experiment()

#             @exp.member
#             class Demo(al.Page):
#                 name = "demo"

#                 def on_exp_access(self):
#                     self += al.MultipleChoiceList("choi1", "choi2", "choi3", name="sel1")

#     """

#     # Documented at :class:`.Element`
#     element_template = jinja_env.get_template("html/SelectElement.html.j2")

#     # Documented at :class:`.SingleChoice`
#     type = "multiple"

#     def __init__(self, *choice_labels, size: int = None, **kwargs):
#         super().__init__(*choice_labels, **kwargs)
#         self.size = size

#     @property
#     def template_data(self):

#         d = super().template_data
#         d["size"] = self.size
#         return d

#     def set_data(self, d):

#         name_map = {str(choice.value): choice.name for choice in self.choices}
#         val = d.get(self.name, None)
#         if not val:
#             return
#         val_name = name_map[val]

#         for i, choice in enumerate(self.choices, start=1):
#             if choice.name == val_name:
#                 self.input[f"choice{i}"] = True
#             else:
#                 self.input[f"choice{i}"] = False


@inherit_kwargs
class SingleChoiceButtons(SingleChoice):
    """
    A prettier :class:`.SingleChoice` element with buttons instead of
    radio inputs.

    Args:
        *choice_labels: Variable numbers of choice labels. See
            :class:`.ChoiceElement` for details.

        button_width: Can be used to manually define the width of
            buttons. If you supply a single string, the same width will
            be applied to all buttons in the element. If you supply
            "auto", button width will be determined automatically. You
            can also supply a list of specific widths for each
            individual button. You must specify a unit, e.g. '140px'.
            Defaults to "equal".

        button_style: Can be used for quick color-styling, using
            Bootstraps default color keywords: btn-primary, btn-secondary,
            btn-success, btn-info, btn-warning, btn-danger, btn-light,
            btn-dark. You can also use the "outline" variant to get
            outlined buttons (eg. "btn-outline-secondary"). If you
            specify a single string, this style is applied to all
            buttons in the element. If you supply a list, you can define
            individual styles for each button. If you supply a list that
            is shorter than the list of labels, the last style
            will be repeated for remaining buttons. Advanced users can
            supply their own CSS classes for button-styling.

        button_round_corners: A boolean switch to toggle whether buttons
            should be displayed with  rounded corners (*True*).

        {kwargs}

    Notes:

        - This element saves answers in the form of integers, counting
          up from 1. Take a look at the examples to see how to work with
          user input to a SingleChoice element.
        - The *align* parameter does not affect the alignment of choice
          labels.

    See Also:

        - :class:`.SingleChoiceBar` is a version of SingleChoiceButtons
          that displays the answering options in a connected bar instead
          of single buttons.
        - The class :class:`.SubmittingButtons` is a version of
          SingleChoiceButtons that automatically moves the experiment
          to the next page, once the participant clicks on an answer.
        - :class:`.MultipleChoiceButtons` can be used to allow multiple
          answers.

    Examples:
        A simple SingleChoiceButtons element::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo_page"

                def on_exp_access(self):
                    self += al.SingleChoiceButtons("Yes", "No", name="c1")

         Accessing the input to a SingleChoiceButtons element::

             import alfred3 as al
             exp = al.Experiment()

             @exp.member
             class Demo(al.Page):
                 name = "demo_page"

                 def on_exp_access(self):
                    self += al.SingleChoiceButtons("Yes", "No", toplab="Choose one", name="c1")


             @exp.member
             class Show(al.Page):

                 def on_first_show(self):
                    c1_answer = self.exp.values["c1] # access value
                    self += al.Text(f"Your answer was: {{c1_answer}}")

    """

    element_template = jinja_env.get_template("html/ChoiceButtons.html.j2")

    #: A boolean switch to toggle whether buttons should be layoutet as
    #: a connected toolbar (*True*), or as separate neighbouring buttons
    #: (*False*, default).
    button_toolbar: bool = False

    #: CSS class for the button group
    button_group_class: str = "choice-button-group"

    def __init__(
        self,
        *choice_labels,
        button_width: Union[str, list] = "equal",
        button_style: Union[str, list] = "btn-outline-dark",
        button_round_corners: bool = True,
        **kwargs,
    ):
        super().__init__(*choice_labels, **kwargs)
        self.button_width: Union[str, list] = button_width
        self.button_style: Union[str, list] = button_style
        self.button_round_corners: bool = button_round_corners

    @property
    def button_style(self) -> Union[str, list]:
        """Union[str, list]: See documentation for the initialization argument."""
        return self._button_style

    @button_style.setter
    def button_style(self, value):
        # create list of fitting length, if only string is provided
        if isinstance(value, str):
            self._button_style = [value for x in self.choice_labels]

        # take styles-list if the length fits
        elif isinstance(value, list) and len(value) == len(self.choice_labels):
            self._button_style = value

        # repeat last value, if styles-list is shorter than labels-list
        elif isinstance(value, list) and len(value) < len(self.choice_labels):
            self._button_style = []
            for i, _ in enumerate(self.choice_labels):
                try:
                    self._button_style.append(value[i])
                except IndexError:
                    self._button_style.append(value[-1])

        elif isinstance(value, list) and len(value) > len(self.choice_labels):
            raise ValueError(
                "List of button styles cannot be longer than list of button labels."
            )

    @property
    def template_data(self) -> dict:

        d = super().template_data
        d["button_style"] = self.button_style
        d["button_group_class"] = self.button_group_class
        d["align_raw"] = self._convert_alignment()
        return d

    def _button_width(self):
        """Add css for button width."""

        if self.button_width == "equal":
            if not self.vertical:
                # set button width to small value, because they will grow to fit the group
                css = f".btn.choice-button-{self.name} {{width: 10px;}} "
            else:
                css = []
            # full-width buttons on small screens
            css += (
                f"@media (max-width: 576px) {{.btn.choice-button-{self.name} {{width:"
                " 100%;}} "
            )
            self._css_code += [(7, css)]

        elif isinstance(self.button_width, str):
            # the group needs to be switched to growing with its member buttons
            css = f"#{self.name} {{width: auto;}} "
            # and return to 100% with on small screens
            css += (
                f"@media (max-width: 576px) {{#{self.name} {{width:"
                " 100%!important;}} "
            )

            # now the width of the individual button has an effect
            css += f".btn.choice-button-{self.name} {{width: {self.button_width};}} "
            # and it, too returns to full width on small screens
            css += (
                f"@media (max-width: 576px) {{.btn.choice-button-{self.name} {{width:"
                " 100%;}} "
            )
            self._css_code += [(7, css)]

        elif isinstance(self.button_width, list):
            if not len(self.button_width) == len(self.choices):
                raise ValueError(
                    "Length of list 'button_width' must equal length of list 'choices'."
                )

            # the group needs to be switched to growing with its member buttons
            css = f"#{self.name} {{width: auto;}} "
            # and return to 100% with on small screens
            css += (
                f"@media (max-width: 576px) {{#{self.name} {{width: 100%!important;}}}}"
            )
            self._css_code += [(7, css)]

            # set width for each individual button
            for w, c in zip(self.button_width, self.choices):
                css = f"#{c.label_id} {{width: {w};}} "
                css += (
                    f"@media (max-width: 576px) {{#{c.label_id} {{width:"
                    " 100%!important;}} "
                )
                self._css_code += [(7, css)]

    def _round_corners(self):
        """Adds css for rounded buttons."""

        spec = "border-radius: 1rem;"
        css1 = f"div#{ self.name }.btn-group>label.btn.choice-button {{{spec}}}"
        css2 = (
            f"div#{ self.name }.btn-group-vertical>label.btn.choice-button {{{spec}}}"
        )
        self.add_css(css1)
        self.add_css(css2)

    def _toolbar(self):
        """Adds css for toolbar display instead of separate buttons."""

        not_ = "last", "first"
        margin = "right", "left"

        for exceptn, m in zip(not_, margin):
            n = "0" if m == "right" else "-1px"
            spec = f"margin-{m}: {n}; "
            spec += f"border-top-{m}-radius: 0; "
            spec += f"border-bottom-{m}-radius: 0;"
            css = (
                f"div#{ self.name }.btn-group>.btn.choice-button:not(:{exceptn}-child)"
                f" {{{spec}}}"
            )
            self._css_code += [(7, css)]

    def _convert_alignment(self):
        if self.vertical:
            if self.align == "center":
                return "align-self-center"
            elif self.align == "left":
                return "align-self-start"
            elif self.align == "right":
                return "align-self-end"
        else:
            return ""

    def prepare_web_widget(self):

        super().prepare_web_widget()

        if self.button_toolbar:
            self._toolbar()

        if self.button_round_corners:
            self._round_corners()

        if not self.button_width == "auto":
            self._button_width()


@inherit_kwargs
class SingleChoiceBar(SingleChoiceButtons):
    """
    A variation of :class:`.SingleChoiceButtons`, which is displayed as
    a toolbar of connected buttons.

    Args:
        *choice_labels: Variable numbers of choice labels. See
            :class:`.ChoiceElement` for details.
        {kwargs}

    Notes:

        - This element saves answers in the form of integers, counting
          up from 1. Take a look at the examples to see how to work with
          user input to a SingleChoice element.
        - The *align* parameter does not affect the alignment of choice
          labels.

    See Also:

        - :class:`.SingleChoiceButtons` is a version of a SingleChoiceBar
          that displays the options in the form of single buttons instead
          of a connected bar.
        - The class :class:`.SubmittingButtons` is a version of
          SingleChoiceButtons that automatically moves the experiment
          to the next page, once the participant clicks on an answer.
        - :class:`.MultipleChoiceButtons` can be used to allow multiple
          answers.

    Examples:
        A simple SingleChoiceBar element::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo_page"

                def on_exp_access(self):
                    self += al.SingleChoiceBar("Yes", "No", name="c1")

       Accessing the input to a SingleChoiceBar element::

            @exp.member
            class Demo(al.Page):
                name = "demo_page"

                def on_exp_access(self):
                    self += al.SingleChoiceBar("Yes", "No", name="c1")

            @exp.member
            class Show(al.Page):

                def on_first_show(self):
                    c1_answer = self.exp.values["c1] # access value
                    self += al.Text(f"Your answer was: {{c1_answer}}")
    """

    # Documented at :class:`.SingleChoiceButtons`
    button_group_class = "choice-button-bar"

    # Documented at :class:`.SingleChoiceButtons`
    button_toolbar = True


@inherit_kwargs
class MultipleChoiceButtons(MultipleChoice, SingleChoiceButtons):
    """
    A prettier :class:`.MultipleChoice` element with buttons instead of
    checkbox inputs.

    To make them distinct from single choice buttons, multiple choice
    buttons don't have rounded corners by default.

    Args:
        *choice_labels: Variable numbers of choice labels. See
            :class:`.ChoiceElement` for details.
        min, max: Minimum (maximum) number of choices that need to be
            selected, if any are selected (does not imply
            *force_input=True*).
        select_hint: Hint to be displayed, if the requirement of
            minimum or maximum number of selected fields is not reached.
            Defaults to the experiment-wide value specified in
            config.conf.
        default: Can be a single integer, or a list of integers which
            indicate the choices that should be selected by default.
            Counting starts at 1.

        {kwargs}

    Notes:
        This element saves and returns not a single value, but a
        dictionary of values. Each choice is represented by a key, and
        the corresponding value is *True*, if the choice was selected and
        *False* otherwise.

        The keys are of the form ``choicei``, where ``i`` is a placeholer
        for the number of the choice, i.e. ``choice1`` for the first choice.

    See Also:
        - See :class:`.MultipleChoice` and :class:`.MultipleChoiceBar`
          for nice-looking buttons and a button-bar of multiple choices.
        - See :class:`.MultipleChoiceList` for a a list that allows
          multiple selections.
        - See :class:`.SingleChoice`, :class:`.SingleChoiceButtons`, and
          :class:`.SingleChoiceBar` for single choice elements.
        - See :class:`.SubmittingButtons` for single choice buttons
          that trigger a forward move on click.

    Examples:
        Multiple choice buttons::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.MultipleChoiceButtons("Yes", "No", "Maybe", name="m1")

        Accessing the input to a MultipleChoiceButtons element::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.MultipleChoiceButtons("a", "b", "c", toplab="Choose one or more", name="c1")

            @exp.member
            class Show(al.Page):

                def on_first_show(self):
                    choices = self.exp.values.get("c1")
                    self += al.Text(f"Your answer is saved like this: {{choices}}")

    """

    def __init__(self, *choice_labels, button_round_corners: bool = False, **kwargs):
        super().__init__(
            *choice_labels, button_round_corners=button_round_corners, **kwargs
        )


@inherit_kwargs
class MultipleChoiceBar(MultipleChoiceButtons):
    """
    A variation of :class:`.MultipleChoiceButtons`, which is displayed as
    a toolbar of connected buttons.

    Args:
        *choice_labels: Variable numbers of choice labels. See
            :class:`.ChoiceElement` for details.
        {kwargs}

    Notes:
        This element saves and returns not a single value, but a
        dictionary of values. Each choice is represented by a key, and
        the corresponding value is *True*, if the choice was selected and
        *False* otherwise.

        The keys are of the form ``choicei``, where ``i`` is a placeholer
        for the number of the choice, i.e. ``choice1`` for the first choice.

    See Also:
        - See :class:`.MultipleChoice` and :class:`.MultipleChoiceButtons`
          for nice-looking buttons and a button-bar of multiple choices.
        - See :class:`.SingleChoice`, :class:`.SingleChoiceButtons`, and
          :class:`.SingleChoiceBar` for single choice elements.
        - See :class:`.SubmittingButtons` for single choice buttons
          that trigger a forward move on click.

    Examples:
        A multiple choice bar with three options::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo_page"

                def on_exp_access(self):
                    self += al.MultipleChoiceBar("Yes", "No", "Maybe", name="m1")

        Accessing the input to a MultipleChoiceBar element::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.MultipleChoiceBar("a", "b", "c", toplab="Choose one or more", name="c1")

            @exp.member
            class Show(al.Page):

                def on_first_show(self):
                    choices = self.exp.values.get("c1")
                    self += al.Text(f"Your answer is saved like this: {{choices}}")

    """

    # Documented at :class:`.SingleChoiceButtons
    button_group_class: str = "choice-button-bar"

    # Documented at :class:`.SingleChoiceButtons
    button_toolbar: bool = True

    # Documented at :class:`.SingleChoiceButtons
    button_round_corners: bool = False


@inherit_kwargs(exclude=["*choice_labels"])
class SelectPageList(SingleChoiceList):
    """
    A :class:`.SingleChoiceList`, automatically filled with page names.

    Args:
        scope: Can be 'exp', or a section name. If *scope* is 'exp', all
            pages in the experiment are listed as choices. If *scope* is
            a section name, all pages in that section are listed.
        include_self: Whether to include the current page in the list,
            if it is in scope.
        check_jumpto: If *True* (default), pages that cannot be jumped
            to will be marked as disabled options. The evaluation is
            based on the :attr:`.Section.allow_jumpto` attribute of
            each page's direct parent section (and only that section).
        check_jumpfrom: If *True*, the element will check, if the
            current page can be left via jump. If not, all pages in the
            list will be marked as disabled options.
        show_all_in_scope: If *True* (default), all pages in the scope
            will be shown, including those that cannot be jumped to.
        display_page_name: If *True*, the page name will be displayed
            in the select list. Defaults to *True*.
        {kwargs}

    Notes:
        This is mostly a utility element for the implementation of
        :class:`.JumpList`.

    Examples:

        Minimal example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo1(al.Page):
                name = "demo1"

                def on_exp_access(self):
                    self += al.SelectPageList(name="select_page")

            @exp.member()
            class Target(al.Page):
                name = "demo2"

    """

    def __init__(
        self,
        scope: str = "exp",
        include_self: bool = False,
        check_jumpto: bool = True,
        check_jumpfrom: bool = True,
        show_all_in_scope: bool = True,
        display_page_name: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.scope = scope
        self.check_jumpto = check_jumpto
        self.check_jumpfrom = check_jumpfrom
        self.show_all_in_scope = show_all_in_scope
        self.include_self = include_self
        self.display_page_name = display_page_name

    def _determine_scope(self) -> List[str]:
        """
        Determines, which pages belong to the scope of the element *and*
        should appear in the dropdown.

        Returns:
            List[str]: List of page names that are in the elements scope
                and should be shown in the dropdown.
        """

        if self.scope in ["experiment", "exp"]:
            scope = list(
                self.experiment.root_section.members["_content"].all_pages.values()
            )
        else:
            try:
                target_section = self.experiment.root_section.all_subsections[
                    self.scope
                ]
                scope = list(target_section.all_pages.values())
            except AttributeError:
                raise AlfredError("Parameter 'scope' must be a section name or 'exp'.")

        choice_labels = []
        if not self.show_all_in_scope:
            for page in scope:
                if page.section.allow_jumpto and page.should_be_shown:
                    choice_labels.append(page.name)
        else:
            choice_labels = [page.name for page in scope]

        if not self.include_self:
            try:
                choice_labels.remove(self.name)
            except ValueError:
                self.log.debug(
                    "ValueError ignored in PageList while trying to remove self from"
                    " the list of choice labels."
                )
                pass

        return choice_labels

    def define_choices(self) -> List[_Choice]:

        choices = []
        for i, page_name in enumerate(self.choice_labels, start=1):
            choice = _Choice()

            choice.label = self._choice_label(page_name)
            choice.type = "radio"
            choice.value = page_name
            choice.name = self.name
            choice.id = f"choice{i}-{self.name}"
            choice.label_id = f"{choice.id}-lab"
            choice.disabled = True if self.disabled else self._jump_forbidden(page_name)
            choice.checked = self._determine_check(i) if not choice.disabled else False
            choice.css_class = f"choice-button choice-button-{self.name}"

            choices.append(choice)

        return choices

    def _jump_forbidden(self, page_name: str) -> bool:
        """
        Returns:
            bool: True, if jumping to the target page or from the
            current page is forbidden. Also True, if the target page
            should not be shown.
        """
        target_page = self.experiment.root_section.all_pages[page_name]

        conditions = [True]  # list of conditions. True means that jumping is allowed

        # disable choice if the target page can't be jumped to
        if self.check_jumpto:
            jump_allowed = target_page.section.allow_jumpto
            conditions.append(jump_allowed)

        # disable choice if self can't be jumped from
        if self.check_jumpfrom:
            jump_allowed = self.page.section.allow_jumpfrom
            conditions.append(jump_allowed)

        conditions.append(target_page.should_be_shown)

        return not all(conditions)

    def _choice_label(self, page_name: str) -> str:
        """
        Returns:
            str: A shortened version of the *page* name, if its length
            exceeds 35 characters.
        """
        target_page = self.experiment.root_section.all_pages[page_name]
        # shorten page title for nicer display
        page_title = target_page.title
        if len(page_title) > 35:
            page_title = page_title[:35] + "..."
        pname = f" (name='{page_name}')" if self.display_page_name else ""
        return f"{page_title}" + pname

    def _determine_check(self, i: int) -> bool:
        """
        For a given choice option, this method determines, whether it
        should be marked as checked by default.

        If the experiment is started in debug mode, the current page is
        marked as checked.

        Returns:
            bool: *True*, if the choice option with index *i* should be
            checked.

        """

        # set default value
        if self.default == i:
            checked = True
        elif self.input == self.choice_labels[i - 1]:
            checked = True
        elif self.debug_enabled and i == 1:
            checked = True
        else:
            checked = False

        if self.experiment.config.getboolean("general", "debug"):
            current_page = self.experiment.movement_manager.current_page
            checked = i == (self.choice_labels.index(current_page.name) + 1)

        return checked

    def prepare_web_widget(self):

        self.choice_labels = self._determine_scope()
        self.choices = self.define_choices()


@inherit_kwargs
class HiddenInput(InputElement):
    """
    Provides a hidden entry field.

    Args:
        {kwargs}

    Examples:
        Minimum example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.HiddenInput(name="hi1", default="fixed")

    """

    base_template = jinja_env.get_template("html/EmptyElement.html.j2")
    element_template = jinja_env.get_template("html/HiddenInputElement.html.j2")


@inherit_kwargs
class DateEntry(InputElement):
    """
    An element for date entry.

    Args:
        default (str): Default date, provided as a string in YYYY-MM-DD
            format. If *None* (default), the current date will be used.
        min (str): Minimum date, provided as a string in YYYY-MM-DD
            format.
        max (str): Maximum date, provided as a string in YYYY-MM-DD
            format.
        {kwargs}

    Examples:
        Minimum example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.DateEntry(default="2022-01-17", name="demo_date")
    """

    element_template = jinja_env.get_template("html/DateEntryElement.html.j2")

    def __init__(self, default: str = None, min: str = None, max: str = None, **kwargs):
        super().__init__(**kwargs)
        self.default = self.validate_date(default) if default is not None else ""
        self.min = self.validate_date(min) if min is not None else ""
        self.max = self.validate_date(max) if max is not None else ""

    @staticmethod
    def validate_date(date: str):
        datetime.strptime(date, "%Y-%m-%d")
        return date

    @property
    def template_data(self):
        d = super().template_data
        d["min"] = self.min
        d["max"] = self.max
        d["default"] = self.default
        return d


@inherit_kwargs
class TimeEntry(InputElement):
    """
    An element for time entry.

    Args:
        default (str): Default time, provided as a string in HH:MM
            format. If *None* (default), the element will display a
            greyed-out time of "12:30".
        min (str): Minimum time, provided as a string in HH:MM
            format.
        max (str): Maximum time, provided as a string in HH:MM
            format.
        {kwargs}

    Examples:
        Minimum example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                def on_exp_access(self):
                    self += al.TimeEntry(default="14:30", name="demo_time")
    """

    element_template = jinja_env.get_template("html/TimeEntryElement.html.j2")

    def __init__(self, default: str = None, min: str = None, max: str = None, **kwargs):
        super().__init__(**kwargs)
        self.default = self.validate_time(default) if default is not None else ""
        self.min = self.validate_time(min) if min is not None else ""
        self.max = self.validate_time(max) if max is not None else ""

    @staticmethod
    def validate_time(time: str):
        datetime.strptime(time, "%H:%M")
        return time

    @property
    def template_data(self):
        d = super().template_data
        d["min"] = self.min
        d["max"] = self.max
        d["default"] = self.default
        return d



"""
Legacy condition module (v2.1.6 and earlier).
"""

import json
import random
import time
import warnings
from collections import Counter
from dataclasses import asdict, dataclass, field
from typing import Iterator, List, Tuple

from ..data_manager import get_data_of_session


class AllConditionsFull(Exception):
    pass


class ConditionInconsistency(Exception):
    pass


@dataclass
class _Session:
    id: str
    timestamp: float = field(default_factory=time.time)

    def active(self, exp) -> bool:
        d = get_data_of_session(exp, self.id)

        if not d:  # if we do not find data, we assume that exp has not been started
            d = exp.db_misc.find_one({"type": "match_group", "group_id": self.id})
            if not d:
                return False

            session_active = []
            for sid in d["roles"].values():
                if sid:
                    active = self.session_active(exp, sid)
                    session_active.append(active)

            active = all(session_active) and len(session_active) > 0

            return active

        return self.check_data(exp, d)

    def session_active(self, exp, sid: str) -> bool:
        data = get_data_of_session(exp, sid)
        return self.check_data(exp, data)

    def check_data(self, exp, data: dict) -> bool:
        aborted = data["exp_aborted"]
        finished = data["exp_finished"]

        if not data.get("exp_start_time", None):  # not started
            return False

        expired = (time.time() - data["exp_start_time"]) > exp.session_timeout
        active = not aborted and not finished and not expired

        return active


@dataclass
class _Slot:

    condition: str
    timeout: int
    sessions: list
    finished: bool = False

    def __init__(self, *sessions, condition: str, timeout: int, finished: bool = False):
        self.condition = condition
        self.timeout = timeout
        self.sessions = [_Session(**s) for s in sessions]
        self.finished = finished

    def status(self, exp):

        if self.finished:
            return "finished"

        elif not self.active_sessions(exp):
            return "free"

        elif not self.finished:
            return "pending"

    def active_sessions(self, exp):
        """
        Returns the active sessions belonging to this slot.
        """
        return [s for s in self.sessions if s.active(exp)]

    @property
    def ids(self):
        return [session.id for session in self.sessions]

    def session(self, id) -> _Session:
        return [s for s in self.sessions if s.id == id]


@dataclass
class _SlotList:
    slots: list

    def __init__(self, *slots):
        self.slots = [_Slot(*s.pop("sessions", []), **s) for s in slots]

    def open_slots(self, exp) -> Iterator[_Slot]:
        return (slot for slot in self.slots if slot.status(exp) == "free")

    def pending_slots(self, exp) -> Iterator[_Slot]:
        return (slot for slot in self.slots if slot.status(exp) == "pending")

    def id_assigned_to(self, id) -> _Slot:
        for slot in self.slots:
            if id in slot.ids:
                return slot
        return None


class _ConditionIO:
    """
    Handles data in- and output for condition administration.
    """

    def __init__(self, exp, respect_version: bool):
        self.exp = exp
        self.version = self.exp.version if respect_version else ""
        self.path = None
        self.query = None
        self.method = None

        if self.exp.secrets.getboolean("mongo_saving_agent", "use"):
            self.method = "mongo"
            self.query = {
                "exp_id": self.exp.exp_id,
                "exp_version": self.version,
                "type": "condition_data",
            }

        elif self.exp.config.get("local_saving_agent", "use"):
            self.method = "local"
            condition_path = self.exp.config.get(
                "exp_condition", "path", fallback="save"
            )
            full_condition_path = self.exp.subpath(condition_path)
            self.path = full_condition_path / f"randomization{self.version}.json"
            self.path.parent.mkdir(exist_ok=True, parents=True)

    def load(self, atomic: bool = True) -> dict:
        if self.method == "mongo":
            if atomic:
                # this will try a couple of times until if receives a version of the data that
                # can be safely worked on (with no other assignment ongoing)
                data = None
                i = 0
                while not data:
                    query = {**self.query, **{"assignment_ongoing": False}}
                    data = self.exp.db_misc.find_one_and_update(
                        query, {"$set": {"assignment_ongoing": True}}
                    )
                    time.sleep(1)
                    i += 1
                    if i > 10:
                        self.exp.log.error(
                            "Could not find a free condition dataset in 10 trys."
                        )
                        break
                return data
            else:

                # if there is no data, this returns None and places an assignment_ongoing note in the DB
                # if there is data, this returns the data and places an assignment_ongoing note in the DB
                data = self.exp.db_misc.find_one_and_update(
                    self.query, {"$set": {"assignment_ongoing": True}}, upsert=True
                )
                return data

        elif self.method == "local":
            try:
                with open(self.path, encoding="utf-8") as f:
                    return json.load(f)
            except FileNotFoundError:
                return None

    def write(self, data: dict, update: bool = False):
        if self.method == "mongo":
            if update:
                # this will update the slots, without touching the assignment status
                query = self.query
                data = {"slots": data["slots"]}
                self.exp.db_misc.find_one_and_update(query, {"$set": data})
            else:
                # this will replace the document, usually leading to releasing the assignment status
                query = {**self.query, **{"assignment_ongoing": True}}
                self.exp.db_misc.find_one_and_replace(query, data, upsert=True)

        elif self.method == "local":
            with open(self.path, "w", encoding="utf-8") as f:
                json.dump(data, f, indent=4)

    def abort(self):
        if self.method == "mongo":
            query = {**self.query, **{"assignment_ongoing": True}}
            self.exp.db_misc.find_one_and_update(
                query, {"$set": {"assignment_ongoing": False}}
            )


class ListRandomizer:
    """
    Offers efficient list randomization.
    Args:
        *conditions: A variable number of tuples, defining the experiment
            conditions. The tuples have the form ``("str", int)``, where
            *str* is a string giving the condition's name, and *int* is
            the target number of observations for this condition.

        exp (alfred3.ExperimentSession): The alfred3 experiment session.

        id (str): A unique identifier for the condition slot assigned
            by this instance of *ListRandomizer*. If *None*, the current
            :attr:`.ExperimentSession.session_id` will be used.
            This argument enables you to assign, for example, a group
            id that connects several sessions, to the randomizer.
            Defaults to None.

        respect_version (bool): If True, randomization will start anew
            for each experiment version. This is especially important,
            if you make changes to the condition setup of an ongoing
            experiment. This might cause the randomizer to fail, causing
            :class:`.ConditionInconsistency` errors.
            Setting *respect_version* to True can fix such issues.
            Defaults to True.

        mode (str): Can be one of 'strict' or 'inclusive'.

            If 'strict', the randomizer will only assign a condition
            slot, if there are no active sessions for that slot. It will
            not assign a condition slot, if a session in that slot
            is finished, or if there is an ongoing session in that slot
            that has not yet timed out. You will end up with exactly
            as many participants in each condition, as specified in the
            target size.

            If 'inclusive', the randomizer will assign a condition slot,
            if there is no finished session in that slot. That means,
            there might be two ongoing sessions for the same slot, and
            both might end up to finish.
            While 'strict' mode might lead to participants being turned
            away before the experiment is really complete, 'inclusive'
            mode might lead to more data being collected than necessary.
            Defaults to 'strict'.

        random_seed: The random seed used for reproducible pseudo-random
            behavior. This seed will be used for shuffling the condition
            list. Defaults to the current system time. Valid seeds are
            all values that are accepted by :func:`random.seed`

        abort_page (alfred3.page.Page): You can reference a custom
                page to be displayed to new participants, if the
                experiment is full.
    The ListRandomizer is used by initializing it (either directly
    or via the convenience method :meth:`.balanced`) and using the
    method :meth:`.get_condition` to receive a condition. By default,
    the ListRandomizer will automatically abort new sessions and display
    an information page for participants, if the experiment is full. This
    behavior can be customized (see :meth:`.get_condition`).
    .. warning:: Be mindful of the argument *respect_version*! With the
        default setting (True), randomization starts from scratch for
        every experiment version. If you set it to False, you will
        run into an error, if you change anything about the conditions.

    Notes:
        **List Randomization**
        In naive randomization, you might end up with a very unbalanced
        design. For example, if you recruit 300 participants and randomize
        them into two conditions, you might end up with 100 participants in
        the first and 200 in the second condition. If you aim for a balanced
        design with 150 participants in either condition, you might have
        to recruit more than 300 participants and throw away lots of observations
        in the condition that turned out to be larger.
        List randomization solves this problem, which is why it is commonly
        used in offline studies. Let's take an easy example. We might have
        two conditions, a and b, each of which should be completed by three participants.
        The order in which participants are assigned to a condition should be
        random. To achieve this, we create a list that contains a condition
        slot for each participant::
            ["a", "a", "a", "b", "b", "b"]
        We then shuffle this list, which might lead to something like the
        following::
            ["b", "a", "a", "b", "a", "b"]

        Then, the conditions are assigned to participants based on the
        shuffled list in the order that they start the experiment.
        **Using the ListRandomizer offline**
        You can use the ListRandomizer offline, i.e. without using a
        mongoDB for data saving. In this case, you must keep in mind,
        that the shuffled list will only be shared on one machine.
        Offline, randomization cannot be synchronized across multiple
        machines.

    Examples:
        A minimal experiment with two conditions. If the experiment
        is full, the experiment will immediately abort new sessions
        and display an abort page to new participants::
            import alfred3 as al
            exp = al.Experiment()
            @exp.setup
            def setup(exp):
                randomizer = al.ListRandomizer(("cond1", 10), ("cond2", 10), exp=exp)
                exp.condition = randomizer.get_condition()

            @exp.member
            class DemoPage(al.Page):
                def on_exp_access(self):
                    if self.exp.condition == "cond1":
                        lab = "label in condition 1"

                    elif self.exp.condition == "cond2":
                        lab = "label in condition 2"

                    self += al.TextEntry(leftlab=lab, name="t1")

        You can use the alternative constructor :meth:`.balanced` for
        simplified initialization, if you use the same sample size for
        all experiment conditions::
            import alfred3 as al
            exp = al.Experiment()
            @exp.setup
            def setup(exp):
                randomizer = al.ListRandomizer.balanced("cond1", "cond2", n=10, exp=exp)
                exp.condition = randomizer.get_condition()

            @exp.member
            class DemoPage(al.Page):
                def on_exp_access(self):
                    if self.exp.condition == "cond1":
                        lab = "label in condition 1"

                    elif self.exp.condition == "cond2":
                        lab = "label in condition 2"

                    self += al.TextEntry(leftlab=lab, name="t1")
    """

    def __init__(
        self,
        *conditions: Tuple[str, int],
        exp,
        id: str = None,
        respect_version: bool = True,
        mode: str = "strict",
        random_seed=None,
        abort_page=None,
        **kwargs,
    ):
        self.exp = exp
        self.id = id if id is not None else exp.session_id
        self.exp.finish_functions.append(self._mark_slot_finished)
        self.respect_version = respect_version
        self.mode = mode
        self.timeout = self.exp.session_timeout
        self.random_seed = random_seed if random_seed is not None else time.time()
        self.conditions = conditions
        self.abort_page = abort_page

        self.io = _ConditionIO(self.exp, respect_version)
        self.slotlist = None

        msg = (
            "The ListRandomizer is operating in compatibility mode. Please switch to"
            " the updated version by using the 'session_ids' parameter instead of 'id'"
            " and increasing your experiment version."
        )
        self.exp.log.warning(msg)
        warnings.warn(msg, FutureWarning)

        if not respect_version:
            raise ValueError(
                "In comptibility mode, the ListRandomizer must be used with"
                " 'respect_version=True'"
            )

        if kwargs:
            msg = (
                "You are using the compatibility version of the ListRandomizer. "
                f"The arguments {kwargs} are not supported and have no effect."
            )

            self.exp.log.warning(msg)

    @property
    def full(self) -> bool:
        """
        bool: Boolean, indicating whether there are any open slots left.
        """
        slot = next(self.slotlist.open_slots(self.exp), None)

        if slot is None and self.mode == "inclusive":
            slot = next(self.slotlist.pending_slots(self.exp), None)

        if slot is not None:
            return False
        else:
            return True

    @classmethod
    def balanced(cls, *conditions, n: int, **kwargs):
        """
        Alternative constructor for the ListRandomizer.
        Args:
            *conditions (str): A variable number of strings, giving the
                condition names.
            n (int): The number of participants **per condition**.
            **kwargs: Keyword arguments, passed on the normal intialization
                of :class:`.ListRandomizer`.

        Notes:
            This alternative constructor offers a more concise syntax
            for simple balanced setups. All conditions receive the same
            sample size.

        Examples:
            ::
                import alfred3 as al
                exp = al.Experiment()
                @exp.setup
                def setup(exp):
                    randomizer = al.ListRandomizer.balanced("cond1", "cond2", n=10, exp=exp)
                    exp.condition = randomizer.get_condition()

                @exp.member
                class DemoPage(al.Page):
                    def on_exp_access(self):
                        if self.exp.condition == "cond1":
                            lab = "label in condition 1"

                        elif self.exp.condition == "cond2":
                            lab = "label in condition 2"

                        self += al.TextEntry(leftlab=lab, name="t1")
        """
        conditions = [(c, n) for c in conditions]
        return cls(*conditions, **kwargs)

    def abort(self):
        self.io.abort()
        full_page_title = "Experiment closed"
        full_page_text = (
            "Sorry, the experiment currently does not accept any further participants."
        )
        self.exp.abort(
            reason="full",
            title=full_page_title,
            msg=full_page_text,
            icon="user-check",
            page=self.abort_page,
        )
        return "__aborted__"

    def get_condition(self, raise_exception: bool = False) -> str:
        """
        Returns a condition.
        Args:
            raise_exception (bool): If True, the function raises
                the :class:`.AllConditionsFull` exception instead of
                automatically aborting the experiment if all conditions
                are full. This allows you to catch the exception and
                customize the experiment's behavior in this case.

        If all conditions are full, the experiment will be aborted. New
        participants will be redirected to an abort page. The text
        displayed on this page can be customized with the arguments
        'full_page_title' and 'full_page_text'.
        Returns:
            str: A condition name, taken from the randomized conditions
            list.

        Raises:
            AllConditionsFull: If raise_exception is True and
            all conditions are full.

        Examples:
            An example using the default behavior::
                import alfred3 as al
                exp = al.Experiment()
                @exp.setup
                def setup(exp):
                    randomizer = al.ListRandomizer(("cond1", 10), ("cond2", 10), exp=exp)
                    exp.condition = randomizer.get_condition()

                @exp.member
                class DemoPage(al.Page):
                    def on_exp_access(self):
                        if self.exp.condition == "cond1":
                            lab = "label in condition 1"

                        elif self.exp.condition == "cond2":
                            lab = "label in condition 2"

                        self += al.TextEntry(leftlab=lab, name="t1")
            This is an example of customizing the 'full' behavior.
            It basically re-implements the default behavior, but showcases how
            you can customize behavior by catching the "AllConditionsFull"
            exception::
                import alfred3 as al
                exp = al.Experiment()
                @exp.setup
                def setup(exp):
                    randomizer = al.ListRandomizer(("cond1", 10), ("cond2", 10), exp=exp)
                    try:
                        exp.condition = randomizer.get_condition(raise_exception=True)
                    except al.AllConditionsFull:
                        full_page = al.Page(title="Experiment closed.", name="fullpage")
                        full_page += al.Text("Sorry, the experiment currently does not accept any further participants")
                        exp.abort(reason="full", page=full_page)


                @exp.member
                class DemoPage(al.Page):
                    def on_exp_access(self):
                        if self.exp.condition == "cond1":
                            lab = "label in condition 1"

                        elif self.exp.condition == "cond2":
                            lab = "label in condition 2"

                        self += al.TextEntry(leftlab=lab, name="t1")

            The following is an example of a custom page displayed when
            the experiment is full. Note that it, basically implements
            the same behavior as the previous example, it is just a little
            more convenient::
                import alfred3 as al
                exp = al.Experiment()
                @exp.setup
                def setup(exp):
                    full_page = al.Page(title="Experiment closed.", name="fullpage")
                    full_page += al.Text("Sorry, the experiment currently does not accept any further participants.")

                    randomizer = al.ListRandomizer(("cond1", 10), ("cond2", 10), exp=exp, abort_page=full_page)

                    exp.condition = randomizer.get_condition()


                @exp.member
                class DemoPage(al.Page):
                    def on_exp_access(self):
                        if self.exp.condition == "cond1":
                            lab = "label in condition 1"

                        elif self.exp.condition == "cond2":
                            lab = "label in condition 2"

                        self += al.TextEntry(leftlab=lab, name="t1")
        """
        self._load_or_insert_data()

        # TODO atomisieren
        assigned_slot = self.slotlist.id_assigned_to(self.id)
        if assigned_slot is not None:
            self.io.write(self._data)
            return assigned_slot.condition

        slot = next(self.slotlist.open_slots(self.exp), None)

        if slot is None and self.mode == "inclusive":
            slot = next(self.slotlist.pending_slots(self.exp), None)

        if slot is None:
            if raise_exception:
                raise AllConditionsFull
            else:
                return self.abort()

        slot.sessions.append(_Session(self.id))
        self.io.write(
            self._data
        )  # releases the assignment, placing the "no assignment ongoing" note
        return slot.condition

    def _validate(self, data):
        if self.respect_version:
            if not self.exp.version == data["exp_version"]:
                raise ConditionInconsistency(
                    "Experiment version and condition slots version do not match."
                )

        data_conditions = [slot["condition"] for slot in data["slots"]]
        counted = Counter(data_conditions)

        instance = dict(self.conditions)

        msg = (
            "Condition data is inconsistent with randomizer specification. You can set"
            " 'respect_version' to True and increase the experiment version."
        )
        if not counted == instance:
            raise ConditionInconsistency(msg)

    def _load_or_insert_data(self):

        # check if there is any data at all, independent of assignment_ongoing status
        data = self.io.load(atomic=False)

        if not data:  # if not, create a dataset and mark it as assignment_ongoing
            self.slotlist = self._randomize()
            data = self._data
            data["assignment_ongoing"] = True
            self.io.write(data)
        elif data and data["assignment_ongoing"]:
            data = (
                self.io.load()
            )  # load data again, this time respecting assignment_ongoing

        self._validate(data)
        self.slotlist = _SlotList(*data["slots"])

        if self.full:
            self.abort()

    def _generate_slots(self) -> List[_Slot]:
        slots = []
        for c in self.conditions:
            slots += [{"condition": c[0], "timeout": self.timeout}] * c[1]
        return slots

    def _randomize(self) -> _SlotList:
        slots = self._generate_slots()
        random.seed(self.random_seed)
        random.shuffle(slots)
        return _SlotList(*slots)

    @property
    def _data(self):
        d = {}
        d["exp_id"] = self.exp.exp_id
        d["exp_version"] = self.exp.version if self.respect_version else ""
        d["type"] = "condition_data"
        d["mode"] = self.mode
        d["slots"] = asdict(self.slotlist)["slots"]
        d["random_seed"] = self.random_seed
        d["assignment_ongoing"] = False
        return d

    def _mark_slot_finished(self, exp):

        interact_data = self.exp.adata.get("interact", {})
        groupid = interact_data.get("groupid", None)
        if groupid is None:
            all_finished = True
        else:
            group_data = exp.db_misc.find_one(
                {"type": "match_group", "group_id": groupid}
            )
            sessions_finished = []
            for sid in group_data["roles"].values():
                if sid == self.exp.session_id:
                    continue
                session_data = exp.db_main.find_one(
                    {"session_id": sid}, projection=["exp_finished"]
                )
                finished = session_data["exp_finished"]
                sessions_finished.append(finished)

            all_finished = all(sessions_finished)

        if not all_finished:
            return

        data = self.io.load(atomic=True)
        self.slotlist = _SlotList(*data["slots"])
        slot = self.slotlist.id_assigned_to(self.id)
        slot.finished = True
        self.io.write(self._data, update=False)


def random_condition(*conditions) -> str:
    """
    Returns a random condition based on the supplied arguments with
    equal probability of all conditions.
    Args:
        *conditions: A variable number of condition identifiers

    Returns:
        str: One of the input arguments, chosen at random.

    See Also:
        This is a naive way of randomizing, suitable mostly for
        quick prototypes with equal probability of all conditions.
        A more powerful approach is offered by :class:`.ListRandomizer`.

    Examples:

            >>> import alfred3 as al
            >>> al.random_condition("A", "B")
            A
        The example returns either "A" or "B".
    """
    return str(random.choice(conditions))



"""
Module for quota functionality.
"""

import json
import random
import time
from dataclasses import asdict, dataclass, field
from pathlib import Path
from traceback import format_exception
from typing import Iterator, List

from pymongo.collection import ReturnDocument

from .data_manager import DataManager, saving_method
from .exceptions import AllSlotsFull, SlotInconsistency


@dataclass
class SessionGroup:

    sessions: List[str]
    aborted_sessions: List[str] = field(default_factory=list)
    expired_sessions: List[str] = field(default_factory=list)

    def query(self, expid) -> dict:
        d = {}
        d["exp_id"] = expid
        d["exp_session_id"] = {"$in": self.sessions}
        d["type"] = DataManager.EXP_DATA

        return d

    def _load_local(self, exp) -> Iterator[dict]:
        dt = DataManager.EXP_DATA
        directory = exp.config.get("local_saving_agent", "path")
        directory = exp.subpath(directory)
        cursor = DataManager.iterate_local_data(dt, directory)
        for data in cursor:
            if data["exp_session_id"] in self.sessions:
                yield data

    def _get_fields(self, exp, fields: List[str]) -> list:
        method = saving_method(exp)
        if "exp_session_id" not in fields:
            fields.append("exp_session_id")
        if method == "mongo":
            data = self._get_fields_mongo(exp, fields)
        elif method == "local":
            data = self._get_fields_local(exp, fields)

        return data

    def _get_fields_mongo(self, exp, fields: List[str]) -> Iterator:
        q = self.query(exp.exp_id)
        projection_fields = {field: 1 for field in fields}
        projection = {**projection_fields, **{"_id": 0}}
        cursor = exp.db_main.find(q, projection=projection)
        return cursor

    def _get_fields_local(self, exp, fields: List[str]) -> Iterator:
        cursor = self._load_local(exp)
        for sessiondata in cursor:
            yield {key: value for key, value in sessiondata.items() if key in fields}

    def _remove_inactive_sessions(self, data: List[dict], move_to: List[str]) -> None:
        for session in data:
            sid = session["exp_session_id"]
            move_to.append(sid)
            try:
                self.sessions.remove(sid)
            except ValueError:
                pass

    def finished(self, exp, data: List[dict] = None) -> bool:
        if not data:
            data = self._get_fields(exp, ["exp_finished"])
        finished = [session["exp_finished"] for session in data]
        return bool(finished) and all(finished)

    def aborted(self, exp, data: List[dict] = None) -> bool:
        if self.aborted_sessions:
            return True

        if not data:
            data = self._get_fields(exp, ["exp_aborted"])

        aborted = []
        aborted_sessions = []
        for session in data:
            aborted.append(session["exp_aborted"])
            if session["exp_aborted"]:
                aborted_sessions.append(session)

        any_aborted = any(aborted)
        if any_aborted:
            self._remove_inactive_sessions(aborted_sessions, self.aborted_sessions)

        return any_aborted

    def expired(self, exp, data: List[dict] = None) -> bool:
        if self.expired_sessions:
            return True

        if not data:
            data = self._get_fields(
                exp, ["exp_start_time", "exp_session_timeout", "exp_save_time"]
            )

        data = list(data)
        now = time.time()

        expired_sessions = []
        for session in data:
            t = session["exp_start_time"]
            timeout = session["exp_session_timeout"]

            if t is None:
                t = session["exp_save_time"]

            passed_time = now - t
            is_expired = passed_time > timeout
            if is_expired:
                expired_sessions.append(session)

        if expired_sessions:
            self._remove_inactive_sessions(expired_sessions, self.expired_sessions)

        return bool(expired_sessions)

    def started(self, exp, data: List[dict] = None) -> bool:
        if not data:
            data = self._get_fields(exp, ["exp_start_time"])
        start_time = [session["exp_start_time"] for session in data]
        if not start_time:
            raise ValueError("Session not found.")
        return not any([t is None for t in start_time])

    def most_recent_save(self, exp, data: List[dict] = None) -> float:
        if not data:
            data = self._get_fields(exp, ["exp_save_time"])

        save_time = [session["exp_save_time"] for session in data]
        if not save_time:
            raise ValueError("Session not found.")
        most_recent = max(save_time)
        return most_recent

    def oldest_save(self, exp, data: List[dict] = None) -> float:
        if not data:
            data = self._get_fields(exp, ["exp_save_time"])

        save_time = [session["exp_save_time"] for session in data]
        if not save_time:
            raise ValueError("Session not found.")
        oldest = min(save_time)
        return oldest

    def pending(self, exp) -> bool:
        fields = [
            "exp_start_time",
            "exp_finished",
            "exp_aborted",
            "exp_save_time",
            "exp_session_timeout",
        ]
        data = list(self._get_fields(exp, fields))

        finished = self.finished(exp, data)
        aborted = self.aborted(exp, data)
        expired = self.expired(exp, data)

        # if not self.started(exp, data):
        #     # use tolerance of 1 min here to give experiments time to start
        #     if time.time() - self.most_recent_save(exp, data) > 60:
        #         return False
        #     else:
        #         return True

        return not finished and not aborted and not expired


@dataclass
class Slot:

    label: str
    session_groups: List[SessionGroup] = field(default_factory=list)
    finished_sessions: List[str] = field(default_factory=list)
    aborted_sessions: List[str] = field(default_factory=list)
    expired_sessions: List[str] = field(default_factory=list)

    def __post_init__(self):
        self.session_groups = [SessionGroup(**sdata) for sdata in self.session_groups]

    def most_recent_save(self, exp) -> float:
        oldest_saves = [s.oldest_save(exp) for s in self.session_groups]
        return max(oldest_saves)

    def npending(self, exp) -> int:
        pending = [s.pending(exp) for s in self.session_groups]
        return len(pending)

    def finished(self, exp) -> bool:
        if self.finished_sessions:
            return True

        session_groups = []
        for s in self.session_groups:
            if s.finished(exp):
                self.finished_sessions.append(s)
            else:
                session_groups.append(s)

        self.session_groups = session_groups

        return bool(self.finished_sessions)

    def pending(self, exp) -> bool:
        pending = []
        for s in self.session_groups:
            if s.finished(exp):
                if s not in self.finished_sessions:
                    self.finished_sessions.append(s)
            elif s.aborted(exp):
                if s not in self.aborted_sessions:
                    self.aborted_sessions.append(s)
            elif s.expired(exp):
                if s not in self.expired_sessions:
                    self.expired_sessions.append(s)
            else:
                pending.append(s)

        self.session_groups = pending
        return bool(pending)

    def conduct_maintenance(self, exp):
        self.pending(exp)

    def open(self, exp) -> bool:
        return not self.finished(exp) and not self.pending(exp)

    def __contains__(self, session_ids: List[str]) -> bool:
        group_contains_session = (
            session_ids == group.sessions for group in self.session_groups
        )
        return any(group_contains_session)


@dataclass
class SlotManager:
    slots: List[dict]

    def __post_init__(self):
        self.slots = [Slot(**slot_data) for slot_data in self.slots]

    def open_slots(self, exp) -> Iterator[Slot]:
        return (slot for slot in self.slots if slot.open(exp))

    def pending_slots(self, exp) -> Iterator[Slot]:
        return (slot for slot in self.slots if slot.pending(exp))

    def find_slot(self, session_ids: List[str]) -> Slot:
        for slot in self.slots:
            if session_ids in slot:
                return slot

    def conduct_maintenance(self, exp):
        [slot.conduct_maintenance(exp) for slot in self.slots]

    def next_pending(self, exp) -> Slot:
        slots = list(self.pending_slots(exp))

        if len(slots) == 1:
            return slots[0]

        slots = self._sparsest_slots(slots, exp)

        if len(slots) == 1:
            return slots[0]

        return self._oldest_slot(slots, exp)

    def _sparsest_slots(self, slots, exp) -> List[Slot]:

        npending = [slot.npending(exp) for slot in slots]
        n = min(npending)
        minimal_pending = [slot for slot in slots if slot.npending(exp) == n]
        return minimal_pending

    def _oldest_slot(self, slots, exp) -> Slot:
        most_recent_save = [slot.most_recent_save(exp) for slot in slots]
        oldest = min(most_recent_save)
        i = most_recent_save.index(oldest)
        return slots[i]


@dataclass
class QuotaData:
    name: str
    exp_id: str
    exp_version: str
    inclusive: bool
    type: str
    slots: List[dict] = field(default_factory=list)
    busy: bool = False
    additional_info: dict = field(default_factory=dict)


class QuotaIO:
    def __init__(self, quota):
        self.quota = quota
        self.exp = quota.exp
        self.db = self.exp.db_misc

        if saving_method(self.exp) == "local":
            self.path.parent.mkdir(exist_ok=True)

    @property
    def query(self) -> dict:
        d = {}
        d["exp_id"] = self.exp.exp_id
        d["exp_version"] = self.quota.exp_version
        d["type"] = self.quota.DATA_TYPE
        d["name"] = self.quota.name
        return d

    @property
    def path(self) -> Path:
        name = f"{self.quota.DATA_TYPE}_{self.quota.name}{self.quota.exp_version}.json"
        directory = self.exp.config.get("data", "save_directory")
        directory = self.exp.subpath(directory)
        return directory / name

    def load(self) -> QuotaData:
        method = saving_method(self.exp)
        if method == "mongo":
            return self.load_mongo(self.quota._insert)
        elif method == "local":
            return self.load_local(self.quota._insert)

    def load_mongo(self, insert: QuotaData) -> QuotaData:
        q = self.query
        data = self.db.find_one_and_update(
            filter=q,
            update={"$setOnInsert": asdict(insert)},
            upsert=True,
            return_document=ReturnDocument.AFTER,
        )
        data.pop("_id", None)
        return QuotaData(**data)

    def load_local(self, insert: QuotaData) -> QuotaData:
        if not self.path.exists():
            self.save_local(asdict(insert))

        else:
            with open(self.path, encoding="utf-8") as fp:
                data = json.load(fp)

            return QuotaData(**data)

    def load_markbusy(self) -> QuotaData:
        method = saving_method(self.exp)
        if method == "mongo":
            return self.load_markbusy_mongo()
        elif method == "local":
            return self.load_markbusy_local()

    def load_markbusy_mongo(self) -> QuotaData:
        q = self.query
        q["busy"] = False

        update = {"$set": {"busy": True}}
        rd = ReturnDocument.AFTER

        data = self.db.find_one_and_update(filter=q, update=update, return_document=rd)

        if data is None:
            return None

        data.pop("_id", None)

        return QuotaData(**data)

    def load_markbusy_local(self) -> QuotaData:
        if not self.path.exists():
            data = asdict(self.rand.data)

        with open(self.path, encoding="utf-8") as fp:
            data = json.load(fp)

        if data["busy"]:
            return None

        data["busy"] = True
        self.save_local(data)
        return QuotaData(**data)

    def save(self, data: QuotaData):
        data = asdict(data)

        method = saving_method(self.exp)
        if method == "mongo":
            self.save_mongo(data)
        elif method == "local":
            self.save_local(data)

    def save_local(self, data: dict):
        with open(self.path, "w", encoding="utf-8") as fp:
            json.dump(data, fp, indent=4)

    def save_mongo(self, data: dict):
        q = self.query
        q["busy"] = True
        self.db.find_one_and_update(filter=q, update={"$set": data})

    def release(self):
        method = saving_method(self.exp)
        if method == "mongo":
            self.release_mongo()
        elif method == "local":
            self.release_local()

    def release_mongo(self):
        q = self.query
        q["busy"] = True
        u = {"$set": {"busy": False}}
        self.db.find_one_and_update(filter=q, update=u)

    def release_local(self):
        with open(self.path, encoding="utf-8") as fp:
            data = json.load(fp)

        data["busy"] = False
        self.save_local(data)

    def __enter__(self):
        data = self.load_markbusy()
        start = time.time()
        wait = 15
        while not data:
            self.exp.log.debug(
                "Could not load non-busy randomizer data. Trying again after waiting"
                " for a short time."
            )
            time.sleep(random.random())
            data = self.load_markbusy()
            if time.time() - start > wait:
                raise RuntimeError(
                    f"Tried to load randomizer data for {wait} seconds. Could not load"
                    " a data, since it was busy."
                )
        return data

    def __exit__(self, exc_type, exc_value, traceback):

        if exc_type and exc_type != AllSlotsFull:
            self.release()
            self.exp.abort(reason="quota_error")
            tb = "".join(format_exception(exc_type, exc_value, traceback))
            self.exp.log.error(
                "There was an error in a locked operation."
                "I aborted the experiment and released the lock."
                f"{tb}"
            )

        else:
            self.release()


class SessionQuota:
    """
    A quota for experiment sessions.

    The quota allows you to enforce an upper limit on the number of
    participants in your experiment.

    Args:
        nslots (int): Maximum number of slots.
        exp (alfred3.ExperimentSession): Experiment session.
        respect_version (bool):
        inclusive (bool):
            If *False* (default), the quota will only assign a
            slot, if there are no pending sessions for that slot. It will
            not assign a slot, if a session in that slot
            is finished, or if there is an ongoing session in that slot
            that has not yet timed out. You will end up with exactly
            as many participants, as specified in *nslots*.

            If *True*, the quota will assign a slot,
            if there is no finished session in that slot. That means,
            there may be two ongoing sessions for the same slot, and
            both might end up to finish.

            While inclusive=*False* may lead to participants being turned
            away before the experiment is really complete, inclusive=True
            may lead to more data being collected than necessary.

            Defaults to *False*.
        name (str):
            An identifier for the quota. If you
            give this a custom value, you can use multiple quotas
            in the same experiment. Defaults to 'quota'.
        abort_page (alfred3.Page): You can reference a custom
                page to be displayed to new participants, if the
                quota is full.

    .. versionadded:: 2.2.0

    Examples:
        A simple example on how to use the quota::

            import alfred3 as al
            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                quota = al.SessionQuota(10, exp)
                quota.count()

            exp += al.Page(title = "Hello, World!", name="hello_world")

    """

    DATA_TYPE = "quota_data"

    def __init__(
        self,
        nslots: int,
        exp,
        respect_version: bool = True,
        inclusive: bool = False,
        name: str = "quota",
        abort_page=None,
    ):
        self.nslots = nslots
        self.slot_label = "slot"
        self.exp = exp
        self.respect_version = respect_version
        self.exp_version = self.exp.version if respect_version else ""
        self.inclusive = inclusive
        self.name = name
        self.session_ids = [exp.session_id]
        self.io = QuotaIO(self)
        self.abort_page = abort_page

        self._initialize_slots()

    @property
    def _insert(self) -> QuotaData:
        data = QuotaData(
            name=self.name,
            exp_id=self.exp.exp_id,
            exp_version=self.exp_version,
            inclusive=self.inclusive,
            type=self.DATA_TYPE,
        )

        return data

    def _initialize_slots(self):
        self.io.load()

        with self.io as data:
            if not data.slots:
                data.slots = self._generate_slots()
                self.io.save(data)

    def _generate_slots(self) -> List[dict]:
        slots = [{"label": self.slot_label}] * self.nslots
        return slots

    @property
    def nopen(self) -> int:
        """
        int: Number of open slots.
        """
        with self.io as data:
            return self._nopen(data)

    def _nopen(self, data) -> int:
        slot_manager = self._slot_manager(data)
        open_slots = list(slot_manager.open_slots(self.exp))
        return len(open_slots)

    @property
    def npending(self) -> int:
        """
        int: Number of slots in which a session is still ongoing.
        """
        with self.io as data:
            return self._npending(data)

    def _npending(self, data) -> int:
        slot_manager = self._slot_manager(data)
        pending_slots = list(slot_manager.pending_slots(self.exp))
        return len(pending_slots)

    @property
    def full(self) -> bool:
        """
        bool: *True*, if the randomizer has allocated all available slots.
        """
        if self.inclusive:
            return (self.nopen + self.npending) == 0
        else:
            return self.nopen == 0

    @property
    def nfinished(self) -> int:
        """
        int: Number of finished slots.
        """
        with self.io as data:
            nopen = self._nopen(data)
            npending = self._npending(data)
            return self.nslots - (nopen + npending)

    @property
    def allfinished(self) -> bool:
        """
        bool: Indicates, whether all slots in the quota are finished.
        """
        return self.nfinished == self.nslots

    def _accepts_sessions(self, data) -> bool:
        """
        bool: *True*, if all slots are full. In 'strict' mode,
        unfinished but pending experiment sessions ('pending' sessions)
        are counted. In 'inclusive' mode, only sessions that are fully
        finished are counted.
        """
        if self.inclusive:
            nopen = self._nopen(data)
            npending = self._npending(data)
            n_unfinished = nopen + npending
            return not n_unfinished == 0

        else:
            return self._nopen(data) > 0

    def count(self, raise_exception: bool = False) -> str:
        """
        Counts the experiment session associated with the quota.

        Args:
            raise_exception (bool): If True, the function raises
                the :class:`.AllSlotsFull` exception instead of
                automatically aborting the experiment if all slots
                are full. This allows you to catch the exception and
                customize the experiment's behavior in this case.

        Returns:
            str: The slot label.

        Raises:
            AllSlotsFull: If raise_exception is True and
            all slots are full.
            SlotInconsistency: If slot validation fails.

        Examples:
            A simple example on how to use the quota::

                import alfred3 as al
                exp = al.Experiment()

                @exp.setup
                def setup(exp):
                    quota = al.SessionQuota(10, exp)
                    quota.count()

                exp += al.Page(title = "Hello, World!", name="hello_world")
        """
        with self.io as data:
            self.exp.log.debug("Loaded quota data. Starting to count.")
            self._validate(data)

            slot_manager = self._slot_manager(data)
            slot = self._own_slot(data)

            if slot:
                self.exp.log.debug(
                    "This session was already assigned to a slot. Returning its slot"
                    " label."
                )
                return slot.label

            full = not self._accepts_sessions(data)
            if full and raise_exception:
                self.exp.log.info(
                    "The quota is full. Aborting count with an exception."
                )
                raise AllSlotsFull
            elif full:
                self.exp.log.info(
                    "The quota is full. Aborting count by aborting the experiment."
                )
                self._abort_exp()
                return "__ABORTED__"

            slot = next(slot_manager.open_slots(self.exp), None)

            if slot is None and self.inclusive:
                self.exp.log.info(
                    "Found no open slot. Searching for a pending slot next, since the"
                    " quota is inclusive."
                )
                slot = slot_manager.next_pending(self.exp)

            if slot is None:
                msg = "No slot found, even though the quota does not appear to be full."
                raise SlotInconsistency(msg)

            self.exp.log.info(
                "The quota found a slot for the current session. Starting to update the"
                " database representations."
            )
            self._update_slot(slot)

            try:  # for compatibility with alfred3-interact
                slot_manager.conduct_maintenance(self.exp)
            except AttributeError:
                pass
            data.slots = asdict(slot_manager)["slots"]
            self.io.save(data)
            self.exp.log.debug(
                "The quota has finished to update the database representations."
                " Returning the slot label now."
            )
            return slot.label

    def _update_slot(self, slot):
        group = SessionGroup(self.session_ids)
        slot.session_groups.append(group)

    def _slot_manager(self, data: QuotaData) -> SlotManager:
        return SlotManager(data.slots)

    def next(self) -> Slot:
        """
        Returns the next open slot.
        """
        with self.io as data:
            open_slots = self._slot_manager(data).open_slots(self.exp)

        return next(open_slots)

    def _own_slot(self, data: QuotaData) -> Slot:
        slot_manager = self._slot_manager(data)
        slot = slot_manager.find_slot(self.session_ids)
        return slot

    def _validate(self, data: QuotaData):
        if self.respect_version:
            if not self.exp.version == data.exp_version:
                raise SlotInconsistency(
                    "Experiment version and randomizer version do not match."
                )

    def _abort_exp(self):
        self.exp.log.info("The quota starts to abort the experiment.")
        full_page_title = "Experiment closed"
        full_page_text = (
            "Sorry, the experiment currently does not accept any further participants."
        )
        self.exp.abort(
            reason="full",
            title=full_page_title,
            msg=full_page_text,
            icon="user-check",
            page=self.abort_page,
        )



"""Run an alfred experiment.

You can either use the command line interface via
``python3 -m alfred3.run`` from within your experiment directory, or
import the `run_experiment` function into your own `run.py` and run it
from there.

Example for importing and running the `run_experiment` function:

.. code-block:: python

    from alfred3.run import ExperimentRunner

    if __name__ == "__main__":
        runner = ExperimentRunner()
        runner.auto_run()

If you want more control over how the app is being run, you can call
the individual methods by hand and call :meth:`ExperimentRunner.app.run`
with arguments of your choice.

.. code-block:: python
    from alfred3.run import ExperimentRunner

    if __name__ == "__main__":
        runner = ExperimentRunner()
        runner.generate_session_id()
        runner.configure_logging()
        runner.create_experiment_app()
        runner.set_port()
        runner.start_browser_thread()
        runner.print_startup_message()
        runner.app.run(use_reloader=False, debug=False)

.. moduleauthor:: Johannes Brachem <jbrachem@posteo.de>
"""

import logging
import platform
import subprocess
import sys
import threading
import webbrowser
from pathlib import Path
from uuid import uuid4

from thesmuggler import smuggle

from alfred3 import alfredlog, localserver
from alfred3._helper import socket_checker
from alfred3.config import ExperimentConfig, ExperimentSecrets


class ExperimentRunner:
    def __init__(self, path: str = None):
        self.expdir = self.find_path(path)
        self.config = ExperimentConfig(self.expdir)
        self.secrets = ExperimentSecrets(self.expdir)
        self.app = None
        self.expurl = None

        self.test_mode = None
        self.debug_mode = None

    def find_path(self, path):
        if path:
            p = Path(path).resolve()
            script0 = p / "script.py"
            if script0.is_file():
                sys.stderr.writelines([f" * Using script '{str(script0)}'\n"])
                return p

        fp = Path(sys.argv[0]).resolve().parent
        script2 = fp / "script.py"
        if script2.is_file():
            sys.stderr.writelines([f" * Using script '{str(script2)}'\n"])
            return fp

        wd = Path.cwd()
        script1 = wd / "script.py"
        if script1.is_file():
            sys.stderr.writelines([f" * Using script '{str(script1)}'\n"])
            return wd

        raise FileNotFoundError("No script.py found.")

    def generate_session_id(self):
        session_id = uuid4().hex
        self.config.read_dict({"metadata": {"session_id": "sid-" + session_id}})

    def configure_logging(self):
        """Sets some sensible logging configuration for local
        experiments.

        * Base logger gets turned off to avoid doubled logging messages
            (we don't want to turn the queue_logger off, because that
            way usage is completely the same between local and web exp.)

        * Queue logger gets configured using settings from config.conf
        """
        config = self.config

        exp_id = config.get("metadata", "exp_id")
        loggername = f"exp.{exp_id}"
        logger = logging.getLogger(loggername)

        formatter = alfredlog.prepare_alfred_formatter(exp_id)

        if config.getboolean("general", "debug"):
            logfile = "alfred_debug.log"
        else:
            logfile = "alfred.log"

        logpath = Path(config.get("log", "path")).resolve() / logfile
        if not logpath.is_absolute():
            logpath = self.expdir / logpath
        file_handler = alfredlog.prepare_file_handler(logpath)
        file_handler.setFormatter(formatter)

        stream_handler = logging.StreamHandler(sys.stderr)
        stream_handler.setFormatter(formatter)

        logger.addHandler(file_handler)
        logger.addHandler(stream_handler)

        lvl = config.get("log", "level")
        if config.getboolean("general", "debug"):
            if config.getboolean("debug", "log_level_override"):
                lvl = config.get("debug", "log_level")

        logger.setLevel(alfredlog.parse_level(lvl))

        base_logger = logging.getLogger("alfred3")
        base_logger.addHandler(logging.NullHandler())

    def create_experiment_app(self):
        script = smuggle(str(self.expdir / "script.py"))

        localserver.Script.expdir = self.expdir
        localserver.Script.config = self.config
        localserver.Script.secrets = self.secrets
        localserver.Script.exp = script.exp

        self.app = localserver.app
        secret_key = self.secrets.get("flask", "secret_key", fallback=None)
        if not secret_key:
            import secrets

            secret_key = secrets.token_urlsafe(16)
        self.app.secret_key = secret_key

        return self.app

    def set_port(self):
        port = 5000
        while not socket_checker(port):
            port += 1
        self.port = port

    def print_startup_message(self):

        msg_startup = (
            f" * Start local experiment using http://127.0.0.1:{self.port}/start\n"
        )
        msg_admin = (
            f" * Start admin mode using http://127.0.0.1:{self.port}/start?admin=true\n"
        )
        msg_test = (
            f" * Start test mode using http://127.0.0.1:{self.port}/start?test=true\n"
        )
        sys.stderr.writelines([msg_startup, msg_admin, msg_test])

    def _open_browser(self):

        # generate url
        expurl = f"http://127.0.0.1:{self.port}/start"
        expurl = expurl + "?" if self.debug_mode or self.test_mode else expurl
        expurl = expurl + "test=true" if self.test_mode else expurl
        expurl = expurl + "debug=true" if self.debug_mode else expurl

        if self.config.getboolean("general", "fullscreen"):
            ChromeKiosk.open(url=expurl)
        else:
            webbrowser.open(url=expurl)

    def start_browser_thread(self):
        # start browser in a thread (needed for windows)
        browser = threading.Thread(target=self._open_browser, name="browser")
        browser.start()

    def auto_run(self, open_browser: bool = None, debug=False, test: bool = False):
        """
        Automatically runs an alfred experiment.

        Args:
            open_browser: Indicates, whether alfred should try to open
                a new browser window automatically.
            debug: Indicates, whether the underlying flask app should be
                run in debug mode. Defaults to None, which leads to
                taking the value from option 'open_browser' in section
                'general' of config.conf.
            test: If true, the experiment is started in test mode.

        """
        self.test_mode = test
        self.debug_mode = debug
        self.generate_session_id()
        self.configure_logging()
        self.create_experiment_app()
        self.set_port()

        open_browser = (
            self.config.getboolean("general", "open_browser")
            if open_browser is None
            else open_browser
        )
        if open_browser:
            self.start_browser_thread()
        self.print_startup_message()
        self.app.run(port=self.port, threaded=False, use_reloader=False, debug=debug)


class ChromeKiosk:
    """Open a Chrome window in kiosk mode."""

    @classmethod
    def open(cls, url: str, path: str = None):
        """Check operating system and call approriate opening method for opening url in Chrome in kiosk mode.

        This will only work, if Chrome is not currently running.

        Args:
            url: URL to open. Needs to start with "http://" or "https://"
            path: Custom path to chrome.exe on Windows. If none is provided, the default paths for Windows 7 and 10 will be tried.
        """
        current_os = platform.system()

        if not url.startswith("http"):
            raise ValueError(
                "Parameter 'url' needs to start with 'http://' or 'https://'."
            )

        if current_os == "Windows":
            cls.open_windows(url=url, path=path)
        elif current_os == "Darwin":
            cls.open_mac(url=url)
        elif current_os == "Linux":
            raise NotImplementedError(
                "This method has not been implemented for Linux distributions."
            )

    @staticmethod
    def open_windows(url: str, path: str = None):
        """Open url in Chrome in kiosk mode on Windows."""

        paths = []
        paths.append(
            Path("C:/Program Files (x86)/Google/Chrome/Application/chrome.exe")
        )
        paths.append(
            Path.home().joinpath("AppData/Local/Google/Chrome/Application/chrome.exe")
        )
        paths.append(Path("C:/Program Files (x86)/Google/Application/chrome.exe"))

        existing_paths = [p for p in paths if p is not None and p.exists()]

        chrome = None

        if path:
            chrome = Path(path)
        else:
            chrome = existing_paths[0]

        if not chrome.exists():
            raise FileNotFoundError(f"Did not find a chrome.exe at {str(chrome)}.")

        subprocess.run([chrome, url, "--kiosk"])

    @staticmethod
    def open_mac(url: str):
        """Open url in Chrome in kisok mode on MacOS."""

        subprocess.run(["open", "-a", "Google Chrome", url, "--args", "--kiosk"])



"""
Provides a command line command for running a local alfred3 experiment.

Run an alfred3 experiment from the terminal by changing to the experiment
directory and running::

    $ alfred3 run

You can access the help for this command directly in the terminal by
executing::

    $ alfred3 run --help

The current version offers the following options::

    Usage: alfred3 run [OPTIONS]

    Options:
    -a, --auto-open / -m, --manual-open
                                    If this flag is set to '-a', the experiment
                                    will open a browser window automatically.
                                    [default: '-a']

    --path TEXT                     Path to experiment directory. [default:
                                    current working directory]

    -debug, --debug / -production, --production
                                    If this flag is set to to '-debug', the
                                    alfred experiment will start in debug mode.
                                    [default: '-production']

    -test, --test / -production, --production
                                    If this flag is set to to '-test', the
                                    alfred experiment will start in test mode.
                                    [default: '-production']

    --help                          Show this message and exit.

"""
from pathlib import Path

import click

from alfred3.run import ExperimentRunner


@click.command()
@click.option(
    "-a/-m",
    "--auto-open/--manual-open",
    default=None,
    help=(
        "If this flag is set to '-a', the experiment will open a browser window"
        " automatically. [default: '-a']"
    ),
)
@click.option(
    "--path",
    default=Path.cwd(),
    help="Path to experiment directory. [default: current working directory]",
)
@click.option(
    "-debug/-production",
    "--debug/--production",
    default=False,
    help=(
        "If this flag is set to to '-debug', the alfred experiment will start in debug"
        " mode. [default: '-production']"
    ),
)
@click.option(
    "-test/-production",
    "--test/--production",
    default=False,
    help=(
        "If this flag is set to to '-test', the alfred experiment will start in test"
        " mode. [default: '-production']"
    ),
)
def run(path, auto_open, debug, test):
    runner = ExperimentRunner(path)
    runner.auto_run(open_browser=auto_open, debug=debug, test=test)



"""
Provides miscellaneous utilities for alfred experiments.

.. versionadded:: 1.5

.. moduleauthor:: Johannes Brachem <jbrachem@posteo.de>

"""

import csv
from pathlib import Path
from typing import Any, Iterator, Tuple, Union

from emoji import emojize

from .element.core import Element, InputElement
from .element.display import Label
from .page import Page


def icon(
    name: str, ml: int = 0, mr: int = 0, size: str = "1rem", spin: bool = False
) -> str:
    """Returns HTML code for displaying font-awesome icons.

    These icons can be used in all places where HTML code is rendered,
    i.e. TextElements, and all labels of elements.

    Args:
        name: The icon name, as shown on https://fontawesome.com/icons?d=gallery&m=free
        ml: Margin to the left, can be an integer from 0 to 5.
        mr: Margin to the right, can be an integer from 0 to 5.

    """
    spin = "fa-spin" if spin else ""
    return (
        f"<i class='fas fa-{name} {spin} ml-{ml} mr-{mr}' style='font-size:"
        f" {size};'></i>"
    )


def emoji(text: str, size: str = "1rem") -> str:
    """Returns a new string in which emoji shortcodes in the input
    string are replaced with their unicode representation.

    Emoji printing can be used in all TextElements and Element labels.
    An overview of shortcodes can be found here:
    https://www.webfx.com/tools/emoji-cheat-sheet/

    Args:
        text: Text, containing emoji shortcodes.

    """
    return f"<span style='font-size: {size};'>{emojize(text, use_aliases=True)}</span>"


def is_section(obj: Any) -> bool:
    """
    Returns True, if the given object is a :class:`.Section`, or a
    subclass of Section.
    """
    from .section import Section

    return isinstance(obj, Section)


def is_page(obj: Any) -> bool:
    """
    Returns True, if the given object is a :class:`.Page`, or a
    subclass of Page.
    """
    return isinstance(obj, Page)


def is_element(obj: Any) -> bool:
    """
    Returns True, if the given object is an :class:`.Element`, or a
    subclass of Element.
    """
    return isinstance(obj, Element)


def is_input_element(obj: Any) -> bool:
    """
    Returns True, if the given object is an :class:`.InputElement`, or a
    subclass of InputElement.
    """
    return isinstance(obj, InputElement)


def is_label(obj: Any) -> bool:
    """
    Returns True, if the given object is an :class:`.Label`, or a
    subclass of Label.
    """
    return isinstance(obj, Label)


def _read_csv_todict(
    path: Union[str, Path], encoding: str = "utf-8", **kwargs
) -> Iterator[dict]:
    with open(path, encoding=encoding) as csvfile:
        reader = csv.DictReader(csvfile, **kwargs)
        yield from reader


def _read_csv_tolist(
    path: Union[str, Path], encoding: str = "utf-8", **kwargs
) -> Iterator[list]:
    with open(path, encoding=encoding) as csvfile:
        reader = csv.reader(csvfile, **kwargs)
        yield from reader


def prefix_keys(d: dict, prefix: str, sep: str = "_") -> dict:
    """
    dict: Returns the input dictionary with prefixed keys.

    Examples:
        >>> a = {"k": "val"}
        >>> prefix_keys(d=a, prefix="demo")
        {"demo_k": "val"}
    """
    keys = [prefix + sep + str(k) for k in d.keys()]
    return {key: val for key, val in zip(keys, d.values())}


def flatten_dict(
    d: dict, prefix_sep: str = "_", sequences_to_dict: bool = True
) -> dict:
    """
    dict: Turns a nested dictionary into a flat one.

    Keys of subdictionaries are concatenated, e.g. ``{"k1": {"s1": "value"}}``
    would result in ``{"k1_s1": "value"}``. The sperator can be defined
    in the argument *prefix_sep*.

    If *sequences_to_dict* is true, values that are iterable sequences
    like lists, tuples or generators (but not strings), will be turned
    into dictionaries and identified with unique keys aswell. In this
    case, the resulting output will be a dictionary where each entry is
    a pair of a single key with a single value.
    """
    out = {}
    for key, val in d.items():

        if isinstance(val, dict):
            renamed = prefix_keys(d=val, prefix=key, sep=prefix_sep)
            flattened = flatten_dict(renamed)

            if any([k in out for k in flattened.keys()]):
                raise ValueError

            out.update(flattened)
        elif isinstance(val, str):
            out[key] = val
        elif sequences_to_dict:
            try:
                dictified_iterable = to_dict(val, prefix=key, sep=prefix_sep)
                if any([k in out for k in dictified_iterable.keys()]):
                    raise ValueError
                out.update(dictified_iterable)
            except TypeError:
                out[key] = val
        else:
            out[key] = val

    return out


def to_dict(data, prefix: str = "", sep: str = "_") -> dict:
    """
    dict: Turns an iterable into a flat dictionary.

    The keys are procuded by counting through the elements of the iterable
    and concatenating them with the *prefix* and *sep*:
    ``key = prefix + sep + str(i)``, where ``i`` is the count.
    """
    out = {}

    for i, val in enumerate(data, 1):
        name = prefix + sep + str(i)

        if isinstance(val, str):
            out[name] = val
        elif isinstance(val, dict):
            renamed = prefix_keys(d=val, prefix=name)
            out.update(flatten_dict(renamed, prefix_sep=sep))
        else:
            try:
                subsequence_dict = to_dict(data=val, prefix=name, sep=sep)
                out.update(subsequence_dict)
            except TypeError:
                out[name] = val

    return out


def prefix_keys_safely(
    data: dict, base: dict, prefix: str = "", sep: str = "_"
) -> dict:
    """
    Takes the *data* dict and prefixes its keys in a way that makes sure
    that there are no keys that are present in both the resulting prefixed
    dictionary and the *base* dictionary.

    If the planned prefixing of any key in *data* would result in a
    conflict with a key in *base*, the separator *sep* will be continually
    repeated until there is no conflict anymore.

    For example, a single underscore would be turned into a double
    underscore on first try. Then into a triple underscore, and so on.

    This can be useful, if you want to update the *base* dictionary with
    the *data* dictionary without the danger of losing data.

    Returns:
        dict: A version of the *data* dictionary, in which all keys received
        the prefix in a way that does not cause conflicts with the *base*
        dictionary.

    Notes:
        Actually, prefixes of the output dictionary are checked very
        conservatively, which makes the process more safe and more efficient. The
        function does not check every single key of the output dictionary,
        but it checks if any key in the *base* dictionary has the same
        prefix. If so, this counts as a key collision and the separator
        will be repeated.

    Examples:

        In this example, prefixing works without any conflict resolution
        being necessary:

        >>> a = {"k": "val1"}
        >>> b = {"k": "val2"}
        >>> prefixed = prefix_keys_safely(data=b, base=a, prefix="demo")
        >>> prefixed
        {"demo_k": "val2"}

        >>> a.update(prefixed)
        >>> a
        {"k": "val1", "demo_k": "val2"}

        Second example, demonstrating how conflicts are resolved:

        >>> a = {"demo_k": "val1"}
        >>> b = {"k": "val2"}
        >>> prefixed = prefix_keys_safely(data=b, base=a, prefix="demo")
        >>> prefixed
        {"demo__k": "val2"}

        >>> a.update(prefixed)
        >>> a
        {"demo_k": "val1", "demo__k": "val2"}

        Third example, demonstrating how convervative the approach is.
        Even though there is no direct collision, the prefix collides
        with the start of a key in ``a``, which causes the function
        to repeat the separator:

        >>> a = {"demo_key1": "val1"}
        >>> b = {"k": "val2"}
        >>> prefixed = prefix_keys_safely(data=b, base=a, prefix="demo")
        >>> prefixed
        {"demo__k": "val2"}

        >>> a.update(prefixed)
        >>> a
        {"demo_key1": "val1", "demo__k": "val2"}
    """
    if not sep:
        raise ValueError("Separator must not be empty.")

    while any([str(k).startswith(prefix) for k in base.keys()]):
        prefix += sep

    return prefix_keys(d=data, prefix=prefix, sep=sep)


def multiple_choice_numbers(choice_dict: dict) -> Union[int, Tuple[int]]:
    """
    Finds the indexes of the choices saved for a multiple choice element.

    Returns:
        tuple: A tuple of integers. The length depends on the number of
            selected choices.
    """
    t = {k: v for k, v in choice_dict.items() if v}
    choice_keys = list(t)
    choice_numbers = [int(key[-1]) for key in choice_keys]

    return tuple(choice_numbers)



"""
.. moduleauthor:: Paul Wiemann <paulwiemann@gmail.com>

_helper contains internal functions which are not to be called by framework users.

"""

import functools
import inspect
import os
import re
import socket
from typing import Union
from urllib.parse import urlparse

from cryptography.fernet import Fernet


def fontsize_converter(font_argument: Union[int, str]) -> str:
    """
    FontsizeConverter checks any font arguments used in alfred and
    returns a fontsize variable compatible with any element or page in
    alfred.

    """
    if font_argument is None:
        return None

    elif font_argument == "normal":
        return "1rem"

    elif font_argument == "big":
        return "1.5rem"

    elif font_argument == "small":
        return "0.75rem"

    elif font_argument == "tiny":
        return "0.5rem"

    elif font_argument == "huge":
        return "2rem"

    elif isinstance(font_argument, int):
        return f"{font_argument}pt"

    else:
        return font_argument


def alignment_converter(alignment_argument, type="text"):
    """
    AlignmentConverter checks any font arguments used in alfred and returns an alignment variable compatible
    for different element types in alfred.

    """

    if type == "text":
        if alignment_argument == "left":
            alignment_argument = "pagination-left"

        elif alignment_argument == "center":
            alignment_argument = "pagination-centered"

        elif alignment_argument == "right":
            alignment_argument = "pagination-right"

    elif type == "container":
        if alignment_argument == "left":
            alignment_argument = "containerpagination-left"

        elif alignment_argument == "center":
            alignment_argument = "containerpagination-centered"

        elif alignment_argument == "right":
            alignment_argument = "containerpagination-right"

    elif type == "both":
        if alignment_argument == "left":
            alignment_argument = "pagination-left containerpagination-left"

        elif alignment_argument == "center":
            alignment_argument = "pagination-centered containerpagination-centered"

        elif alignment_argument == "right":
            alignment_argument = "pagination-right containerpagination-right"

    elif type == "div":
        if alignment_argument == "left":
            alignment_argument = "text-align:left"

        elif alignment_argument == "center":
            alignment_argument = "text-align:center"

        elif alignment_argument == "right":
            alignment_argument = "text-align:right"

    return alignment_argument


class Decrypter:

    _decrypter = None

    def decrypt_login(self, username=None, password=None, from_env=False):

        if not self._decrypter:
            # Fernet instance for decryption of login data
            if os.path.isfile("alfred_secrect.key"):
                with open("alfred_secrect.key", "rb") as keyfile:
                    key = keyfile.read()
            else:
                key = os.environ.get("ALFRED_SECRET_KEY")
            try:
                self._decrypter = Fernet(key)
            except Exception:
                RuntimeError(
                    "Unable to initialize Fernet decrypter: Secret key not found!"
                )

        if from_env:
            try:
                decrypted_username = self._decrypter.decrypt(
                    os.environ.get("ALFRED_MONGODB_USER").encode()
                ).decode()
                decrypted_password = self._decrypter.decrypt(
                    os.environ.get("ALFRED_MONGODB_PASSWORD").encode()
                ).decode()

                return (decrypted_username, decrypted_password)

            except (AttributeError, NameError):
                print(
                    "Incomplete DB login data in environment variables. Now trying to"
                    " decrypt login data from config.conf..."
                )

        decrypted_username = self._decrypter.decrypt(username.encode()).decode()
        decrypted_password = self._decrypter.decrypt(password.encode()).decode()

        return (decrypted_username, decrypted_password)


class _DictObj(dict):
    """
    This class allows dot notation to access dict elements

    Example:
    d = _DictObj()
    d.hello = "Hello World"
    print d.hello # Hello World
    """

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def is_url(url=None):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc, result.path])
    except Exception:
        return False


def check_name(name: str):

    if name in ["exp", "experiment"]:
        raise ValueError(f"{name} cannot be chosen as a name.")

    if not re.match(pattern=r"^[a-zA-z](\d|_|[a-zA-Z])*$", string=name):

        raise ValueError(
            "Name must start with a letter and can include only "
            "letters (a-z, A-Z), digits (0-9), and underscores ('_')."
            f"Name '{name}' does not match this pattern."
        )


def socket_checker(port):
    try:
        s = socket.socket()
        s.bind(("127.0.0.1", port))
        s.listen(1)
        s.close()
        return True
    except Exception:
        s.close()
        return False


def sort_dict(d: dict) -> dict:
    """Returns a dict, sorted alphabetically by its keys."""
    return {key: d[key] for key in sorted(d)}


def add_indent(inp: str, spaces: int = 8) -> str:
    """
    Adds indentation to all lines of a single or multiline string.

    Args:
        inp: Input string
        spaces: Number of spaces to indent each line by
    """
    splitted = inp.split("\n")
    indented = f"\n{' '*spaces}".join(splitted)
    return " " * spaces + indented


def build_table(docs: dict, caption: str = "", widths: str = "20, 80") -> str:
    """
    Transforms a documentation dictionary into a string, representing
    the docutils csv-table directive.

    Args:
        docs: Documentation dictionary
        caption: String to use as table caption
        widths: String, indicating the relative widths of the table
            columns in percent.
    """
    directive = f".. csv-table:: {caption}"
    indent = " " * 3

    rows = []
    rows.append(f":widths: {widths}\n")

    for arg, explanation in docs.items():
        rows.append(f'"{arg}", "{explanation}"')

    built_rows = f"\n{indent}".join(rows)
    return directive + "\n" + indent + built_rows


def build_kwargs(
    docs: dict, text: str = "Inherited keyword arguments\n", **kwargs
) -> str:
    """
    Builds a full keyword-arguments string, ready for use in a class
    docstring.

    Args:
        docs: Documentation dictionary
        text: A short heading for the keyword arguments
        **kwargs: Passed on to :meth:`.build_table`

    """
    heading = f"\\*\\*kwargs: {text}\n"
    body = build_table(docs=docs, **kwargs)
    body = add_indent(body, spaces=12)
    return heading + body


def extract_arguments(obj) -> dict:
    """
    Extracts function arguments from google style docstrings of the
    input object.

    Returns:
        dict: Dictionary of argument names and their descriptions.
    """
    args = {}
    beginning_found = False
    previous_arg = None
    p = re.compile(
        r"    (?P<arg>[\w*]+[\w ,]*?) ?(\((?P<type>.+?)?\))?:(?P<description>.*)"
    )

    for line in inspect.getdoc(obj).split("\n"):
        if line in ["Args:", "Arguments:"]:
            beginning_found = True
            continue

        if beginning_found:
            m = p.match(line)
            if m is not None:
                name = m["arg"] if m["type"] is None else f"{m['arg']} ({m['type']})"
                args[name] = [m["description"]]
                previous_arg = name
            elif line.startswith(" " * 8):
                args[previous_arg].append(line)

            elif beginning_found and line.replace(" ", "") != "":
                break

    def clean(lines: list) -> str:
        """
        All occurances of single and double quotes are removed,
        because otherwise they would mess up the csv table.

        """
        joined = " ".join(lines)
        stripped = joined.strip()
        escaped1 = stripped.replace("'", "")
        escaped2 = escaped1.replace('"', "")
        return re.sub(r" +", " ", escaped2)

    return {arg: clean(desc) for arg, desc in args.items()}


def extract_arguments_from_tree(obj) -> dict:
    args = {}

    mro = list(inspect.getmro(obj))
    mro.reverse()

    for parent in mro:
        parent_args = extract_arguments(parent)
        parent_args.pop("**kwargs", None)
        args.update(parent_args)

    return args


def inherit_kwargs(
    _klass=None,
    *,
    from_: list = None,
    not_from_: list = None,
    include: list = None,
    exclude: list = None,
    sort_kwargs: bool = True,
    build_function: callable = build_kwargs,
    **kwargs,
):
    """
    Decorator for easier docstring sharing, used for displaying
    documentation of inherited keyword arguments.

    Args:
        ``from_`` : List of classes to inherit argument documentation from.
        ``not_from_`` : List of classes *not* to inherit argument
            documentation from. Useful, if you want to include some
            specific parent or grandparent.
        include: List of keyword argument names to include
        exclude: List of keyword argument names to exclude
        sort_kwargs: If *True*, the keyword arguments will be arranged
            in alphabetical order. Defaults to True.
        build_function: A function that, as a minimmum, takes an argument
            of *docs*, which is the dictionary of arguments and their
            documentation. It must return a string. Its output is
            inserted in the decorated class' docstring. The kwargs
            will be passed on to the build_function.
        **kwargs: Further keyword arguments. These will be passed on to
            the *build_function*.

    Notes:
        Replaces the placeholder ``{kwargs}`` in the docstring of the
        decorated object with argument documentation extracted from
        the parent classes.

        Further notes:

        1. The decorator can deal with classes that inherit from more
           than one class.

        2. It is possible to use a different formatting by supplying
           a different build function.

        3. Arguments that are already part of the decorated class'
           docstring will be filtered out.

        4. By using the *from_* argument, you can inherit from a
           completely different class that is not part of the decorated
           class' hierarchy.

        5. If you are using curly braces (``{{}}``) anywhere in your
           docstrings, you have to escape them to let *inherit_kwargs*
           know that they are not a placeholder. Otherwise you will
           receive a somewhat cryptic error message
           ("IndexError: Replacement index 0 out of range for positional
           args tuple").

    Examples:

        A simple example::

            from alfred3._helper import inherit_kwargs

            class Parent:
                '''
                Parent docstring.

                Args:
                    arg1: Description
                    arg2: Description
                '''

                def __init__(self, arg1, arg2):
                    pass


            @inherit_kwargs
            class Child(Parent):
                '''
                Child docstring.

                Args:
                    arg3: Description
                    {kwargs}
                '''

                def __init__(self, arg3, **kwargs):
                    super().__init__(**kwargs)


    """
    exclude = exclude if exclude is not None else []

    def build_kwargs(klass):
        @functools.wraps(klass)
        def wrapper():

            # collect arguments from parent classes
            inherited_docs = {}
            parents = from_ if from_ is not None else klass.__bases__
            for parent in parents:
                if not_from_ is not None and parent not in not_from_:
                    continue
                inherited_docs.update(extract_arguments_from_tree(parent))

            # remove arguments that are defined in klass directly
            klass_args = extract_arguments(klass)
            inherited_docs = {
                k: v for k, v in inherited_docs.items() if k not in klass_args
            }

            # apply inclusion and exclusion
            for arg in list(inherited_docs.keys()):
                for ex in exclude:
                    ex = re.sub(r"\*", r"\*", ex)
                    m = re.match(ex + r"(\s?\(.+\))?", arg)
                    if m:
                        del inherited_docs[arg]
                if include is not None:
                    if not re.sub(r"(\s?\(.+\))?", "", arg) in include:
                        del inherited_docs[arg]

            if sort_kwargs:
                inherited_docs = sort_dict(inherited_docs)

            doc_kwargs = build_function(docs=inherited_docs, **kwargs)

            # replace docstring
            klass_doc = klass.__doc__.format(kwargs=doc_kwargs)
            klass.__doc__ = klass_doc

            return klass

        return wrapper()

    if _klass is None:
        return build_kwargs
    else:
        return build_kwargs(_klass)



from ._version import __version__
from .element.action import (
    BackButton,
    Button,
    DynamicJumpButtons,
    ForwardButton,
    JumpButtons,
    JumpList,
    SubmittingBar,
    SubmittingButtons,
)
from .element.core import Row, RowLayout, Stack
from .element.display import (
    Alert,
    Audio,
    BarLabels,
    ButtonLabels,
    Card,
    CodeBlock,
    CountDown,
    CountUp,
    Hline,
    Html,
    Image,
    Label,
    MatPlot,
    ProgressBar,
    Text,
    VerticalSpace,
    Video,
)

# from .element.input import MultipleChoiceList
from .element.input import (
    DateEntry,
    EmailEntry,
    HiddenInput,
    MatchEntry,
    MultipleChoice,
    MultipleChoiceBar,
    MultipleChoiceButtons,
    NumberEntry,
    PasswordEntry,
    RangeInput,
    RegEntry,
    SelectPageList,
    SingleChoice,
    SingleChoiceBar,
    SingleChoiceButtons,
    SingleChoiceList,
    TextArea,
    TextEntry,
    TimeEntry,
)
from .element.misc import (
    Callback,
    Data,
    HiddenInput,
    HideNavigation,
    JavaScript,
    RepeatedCallback,
    Style,
    Value,
    WebExitEnabler,
)
from .experiment import Experiment
from .page import (
    AutoClosePage,
    AutoForwardPage,
    NoDataPage,
    NoNavigationPage,
    NoSavingPage,
    Page,
    PasswordPage,
    UnlinkedDataPage,
    WidePage,
)
from .quota import SessionQuota
from .randomizer import ListRandomizer, random_condition
from .section import ForwardOnlySection, HideOnForwardSection, RevisitSection, Section
from .util import (
    emoji,
    icon,
    is_element,
    is_input_element,
    is_label,
    is_page,
    is_section,
    multiple_choice_numbers,
)



"""
.. moduleauthor:: Paul Wiemann <paulwiemann@gmail.com>, Johannes Brachem <jbrachem@posteo.de>
"""

import copy
import json
import time
from dataclasses import asdict
from pathlib import Path
from typing import Dict, Iterator, List, Union

from cryptography.fernet import Fernet, InvalidToken

from .alfredlog import QueuedLoggingInterface
from .config import ExperimentSecrets
from .exceptions import AlfredError
from .saving_agent import AutoMongoClient
from .util import flatten_dict, prefix_keys_safely


class DataManager:
    EXP_DATA = "exp_data"
    UNLINKED_DATA = "unlinked"
    CODEBOOK_DATA = "codebook"
    HISTORY = "move_history"

    instance_log = False

    _client_data_keys = {
        "client_resolution_screen",
        "client_resolution_inner",
        "client_referrer",
        "client_javascript_active",
        "client_device_type",
        "client_device_manufacturer",
        "client_device_family",
        "client_browser",
        "client_os_family",
        "client_os_name",
        "client_os_version",
    }

    _metadata_keys = {
        "exp_author",
        "exp_title",
        "exp_version",
        "exp_start_time",
        "exp_start_timestamp",
        "exp_save_time",
        "exp_finished",
        "exp_aborted",
        "exp_aborted_because",
        "exp_session",
        "exp_condition",
        "exp_id",
        "exp_session_id",
        "exp_plugin_queries",
        "exp_session_timeout",
        "session_status",
        "alfred_version",
        "type",
    }

    def __init__(self, experiment):
        self._experiment = experiment
        self.exp = experiment
        self.additional_data = {}
        self.log = QueuedLoggingInterface(base_logger=__name__)
        self.log.add_queue_logger(self, __name__)

        self.client_data = {key: None for key in self._client_data_keys}

    @property
    def experiment(self):
        return self._experiment

    @property
    def metadata(self) -> dict:
        data = {}

        data["exp_author"] = self._experiment.author
        data["exp_title"] = self._experiment.title
        data["exp_version"] = self._experiment.version
        data["exp_start_time"] = self._experiment.start_time
        data["exp_start_timestamp"] = self._experiment.start_timestamp
        data["exp_save_time"] = time.time()
        data["exp_finished"] = self._experiment.finished
        data["exp_aborted"] = self._experiment.aborted
        data["exp_aborted_because"] = self._experiment._aborted_because
        data["exp_session"] = self._experiment.session
        data["exp_condition"] = self._experiment.condition
        data["exp_id"] = self._experiment.exp_id
        data["exp_session_id"] = self._experiment.session_id
        data["exp_plugin_queries"] = self._experiment._plugin_data_queries
        data["exp_session_timeout"] = self._experiment.session_timeout
        data["session_status"] = self._experiment.session_status
        data["alfred_version"] = self._experiment.alfred_version
        data["type"] = self.EXP_DATA

        if not len(self._metadata_keys) == len(data):
            diff = self._metadata_keys ^ set(self._metadata)
            raise AlfredError(
                "Length of metadata key set and keys in metadata dict do not match."
                f" Differences: {diff}"
            )

        return data

    @property
    def move_history(self):
        return [asdict(move) for move in self.exp.movement_manager.history]

    @property
    def values(self):
        return {el["name"]: el["value"] for el in self.exp.root_section.data.values()}

    @property
    def element_data(self):
        eldata = self.experiment.root_section.data
        return {**eldata}

    @property
    def session_data(self):
        d = {**self.metadata, **self.client_data}
        d["exp_data"] = self.element_data
        d["exp_move_history"] = self.move_history
        d["additional_data"] = self.additional_data
        return d

    @property
    def codebook_data(self) -> Dict[str, dict]:
        """dict: Returns codebook data for the current session."""
        exp = self.extract_codebook_data(self.session_data)
        unlinked = self.extract_codebook_data(self.unlinked_data)

        return {**exp, **unlinked}

    @staticmethod
    def extract_codebook_data(exp_data: dict) -> Dict[str, dict]:
        """
        Extracts codebook data from a full experiment data dictionary.

        Args:
            exp_data: Full experiment data dictionary, like the one
                returned by :attr:`.session_data`

        Returns:
            dict: A dictionary of dictionaries, containing detailed
                information about the used elements.
        """
        meta = {}
        meta["alfred_version"] = exp_data["alfred_version"]
        meta["exp_author"] = exp_data["exp_author"]
        meta["exp_title"] = exp_data["exp_title"]
        meta["exp_version"] = exp_data["exp_version"]

        codebook = exp_data.pop("exp_data")
        for entry in codebook.values():
            entry.pop("value", None)
            entry.update(meta)

        return codebook

    @staticmethod
    def extract_fieldnames(data: Iterator) -> list:
        """
        Finds correct fieldnames for exporting multiple unlinked or
        codebook datasets to a single csv file.

        Args:
            data: A list or other iterator, providing the individual
                datasets

        Returns:
            list: List of fieldnames

        Note:
            The first scanned document determines the initial order of
            the fieldnames. Subsequently found additional fieldnames
            are simply appended.
        """
        fieldnames = {}
        for element in data:
            el = copy.copy(element)
            el.pop("_id", None)  # remove mongoDB doc ID, if there is one
            fieldnames.update(el)

        return list(fieldnames)

    @classmethod
    def extract_ordered_fieldnames(cls, data: Iterator) -> list:
        """
        Finds correct (and correctly ordered) fieldnames for exporting
        multiple experiment datasets to a single csv file.

        Args:
            data: A list or other iterator providing the individual
                datasets. These need to be the full experiment datasets.

        Returns:
            list: List of fieldnames

        Notes:
            The fieldnames are ordered as follows:

            1. Experiment metadata
            2. Client info data
            3. Element data (ordered alphabetically by element name)
            4. Additional data

        """
        metadata = []
        client_info = []
        adata = []
        elements = {}

        for dataset in data:
            d = cls.flatten(copy.copy(dataset))

            for entry in list(d.keys()):
                if entry in cls._client_data_keys:
                    d.pop(entry)
                    client_info.append(entry)
                elif entry in cls._metadata_keys:
                    d.pop(entry)
                    metadata.append(entry)
                elif entry.startswith("additional_data"):
                    d.pop(entry)
                    adata.append(entry)
            elements.update(d)

        metadata = [name for name in metadata if not name.startswith("additional_data")]

        element_names = sorted(list(elements))
        metadata = sorted(list(set(metadata)))
        client_info = sorted(list(set(client_info)))
        adata = sorted(list(set(adata)))

        fieldnames = metadata + client_info + element_names + adata
        return fieldnames

    @staticmethod
    def sort_fieldnames(fieldnames: List[str], template: List[str]) -> List[str]:
        """
        Sorts the list fieldnames according to template.

        If fieldnames contains more fields than template, the returned
        list consists of a sorted part (the elements present in
        *template*) and an appended part (the other elements, appended
        in their order of appearance in *fieldnames*).

        Args:
            fieldnames: List to sort
            template: Template list, indicating the desired order of
                the known elements in fieldnames.

        Returns:
            list: Sorted list

        Raises:
            ValueError: If either of the provided lists contain
                duplicate values.
        """
        if len(set(fieldnames)) != len(fieldnames):
            raise ValueError("Input list must contain only unique elements.")

        if len(set(template)) != len(template):
            raise ValueError("Template list must contain only unique elements.")

        def find_position(name):
            try:
                return template.index(name)
            except ValueError:
                return len(template) + 1

        return sorted(fieldnames, key=find_position)

    @classmethod
    def sort_codebook_fieldnames(cls, fieldnames: List[str]) -> List[str]:

        t1 = ["exp_title", "exp_author", "exp_version", "alfred_version"]
        t2 = [
            "element_type",
            "name",
            "label_top",
            "label_left",
            "label_right",
            "label_bottom",
        ]
        t3 = [
            "force_input",
            "default",
            "placeholder",
            "prefix",
            "suffix",
            "description",
        ]

        template = t1 + t2 + t3

        return cls.sort_fieldnames(fieldnames, template)

    @staticmethod
    def flatten(data: dict) -> dict:
        eldata = data.pop("exp_data")
        data.pop("exp_move_history", None)
        data.pop("exp_plugin_queries", None)
        data.pop("_id", None)

        values = {}
        for name, elmnt in eldata.items():
            try:
                # if the value is a dictionary, like in multiple choice elements
                for subname, val in elmnt["value"].items():
                    values[f"{name}_{subname}"] = val
            except AttributeError:
                values[name] = elmnt["value"]

        adata = data.pop("additional_data", {})
        adata = flatten_dict(adata)

        maindata = {**data, **values}
        adata = prefix_keys_safely(data=adata, base=maindata, prefix="additional_data")
        return {**maindata, **adata}

    @staticmethod
    def regularize(data: dict) -> dict:
        pass

    @property
    def unlinked_data(self):
        data = {}
        data["type"] = self.UNLINKED_DATA
        data["exp_data"] = self._experiment.root_section.unlinked_data
        data["exp_author"] = self._experiment.author
        data["exp_title"] = self._experiment.title
        data["exp_id"] = self._experiment.exp_id
        data["exp_version"] = self._experiment.version
        data["alfred_version"] = self._experiment.alfred_version
        return data

    @property
    def unlinked_values(self):
        return {
            el["name"]: el["value"]
            for el in self.exp.root_section.unlinked_data.values()
        }

    def unlinked_data_with(self, saving_agent):
        if saving_agent.encrypt:
            return self.exp.data_manager.encrypt_values(self.unlinked_data)
        else:
            return self.unlinked_data

    @property
    def flat_session_data(self):
        return self.flatten(self.session_data)

    def flat_unlinked_data(self):
        return self.flatten(self.unlinked_data)

    def encrypt_values(self, data: dict) -> dict:
        data = copy.copy(data)
        for eldata in data["exp_data"].values():
            eldata["value"] = self.exp.encrypt(eldata["value"])
        return data

    def decrypt_values(self, data: dict) -> dict:
        data = copy.copy(data)
        for eldata in data["exp_data"].values():
            eldata["value"] = self.exp.decrypt(eldata["value"])
            return data

    def get_page_data(self, name: str) -> dict:
        return self.experiment.root_section.all_pages[name].data

    def get_section_data(self, name: str) -> dict:
        return self.experiment.root_section.all_subsections[name].data

    def iter_flat_mongo_data(self, data_type: str = "exp_data") -> Iterator[dict]:
        """
        Iterates over all datasets saved in the mongoDB associated with
        the experiment via its mongo_saving_agent.

        Args:
            data_type: Can be one of 'exp_data', or 'unlinked_data'

        Yields:
            dict: Flattened experiment data.
        """
        cursor = self.iterate_mongo_data(
            exp_id=self.exp.exp_id, data_type=data_type, secrets=self.exp.secrets
        )

        for dataset in cursor:
            yield self.flatten(dataset)

    def iter_flat_local_data(self, data_type: str = "exp_data") -> Iterator[dict]:
        """
        Iterates over all datasets saved via the experiment's
        local_saving_agent.

        Args:
            data_type: Can be one of 'exp_data', or 'unlinked_data'

        Yields:
            dict: Flattened experiment data
        """
        if data_type == "exp_data":
            path = self.exp.config.get("local_saving_agent", "path")
        elif data_type == "unlinked":
            path = self.exp.config.get("local_saving_agent_unlinked", "path")

        path = self.exp.subpath(path)
        cursor = self.iterate_local_data(data_type=data_type, directory=path)

        for dataset in cursor:
            yield self.flatten(dataset)

    @staticmethod
    def iterate_mongo_data(
        exp_id: str, data_type: str, secrets: ExperimentSecrets, exp_version: str = None
    ) -> Iterator[dict]:
        """Returns a MongoDB cursor, iterating over the experiment
        data in the database.

        .. versionadded:: 1.5

        Usage::
            cursor = data_manager.iterate_mongo_data(data_type="exp_dat")
            for doc in cursor:
                ...

        Args:
            exp_id: Experiment id
            data_type: The type of data to be collected. Can be
                'exp_data', 'codebook', or 'unlinked'.
            secrets: Experiment secrets configuration object. This is
                used to extract information for database access.
            exp_version: If specified, data will only be queried for
                this specific version.
        """
        if data_type != "exp_data":
            section_name = f"mongo_saving_agent_{data_type}"
        else:
            section_name = "mongo_saving_agent"

        section = secrets.combine_sections("mongo_saving_agent", section_name)
        dbname = section["database"]
        colname = section["collection"]

        client = AutoMongoClient(section)
        db = client[dbname][colname]
        query = {"exp_id": exp_id, "type": data_type}

        if exp_version is not None:
            query["exp_version"] = exp_version

        return db.find(query)

    @classmethod
    def iterate_local_data(
        cls,
        data_type: str,
        directory: Union[str, Path],
        exp_version: str = None,
    ) -> Iterator[dict]:
        """Generator function, iterating over experiment data .json files
        in the specified directory.

        .. versionadded:: 1.5

        Usage::
            cursor = data_manager.collect_local_data(data_type="exp_data")
            for doc in cursor:
                ...

        Args:
            data_type: The type of data to be collected. Can be
                'exp_data' or 'unlinked'.
            exp_version: If specified, data will only be queried for
                this specific version.
            directory: The directory in which to look for data.
        """
        path = Path(directory).resolve()
        if not path.is_absolute():
            raise ValueError("directory must be absolute")

        if not path.exists():
            return

        for fp in path.iterdir():
            if not fp.suffix == ".json":
                continue

            try:
                with open(fp, encoding="utf-8") as f:
                    doc = json.load(f)
            except json.decoder.JSONDecodeError:
                continue
            except IsADirectoryError:
                continue

            doctype = doc.get("type")

            if data_type != doctype:
                continue

            yield doc


def decrypt_recursively(
    data: Union[list, dict, int, float, str, bytes], key: bytes
) -> Union[list, dict, int, float, str, bytes]:
    """
    Used mainly for decrypting encrypted values in a nested dictionary.
    Can also decrypt single values.

    The encryption/decryption mechanism is symmetric.

    Returns:
        Decrypted object of the same type as the input.

    """
    f = Fernet(key=key)

    if isinstance(data, bytes):
        try:
            return f.decrypt(data)
        except InvalidToken:
            return data

    if isinstance(data, (int, float, str)):
        try:
            original_type = type(data)
            data_in_bytes = str(data).encode()
            decrypted_value = original_type(f.decrypt(data_in_bytes).decode())
            return decrypted_value
        except InvalidToken:
            return data

    elif isinstance(data, list):
        return [decrypt_recursively(entry, key=key) for entry in data]

    elif isinstance(data, dict):
        return {k: decrypt_recursively(v, key=key) for k, v in data.items()}


def saving_method(exp) -> str:
    if not exp.secrets.getboolean("mongo_saving_agent", "use"):
        if exp.config.getboolean("local_saving_agent", "use"):
            return "local"
    elif exp.secrets.getboolean("mongo_saving_agent", "use"):
        return "mongo"
    else:
        return None


def get_session_mongo(exp, sid) -> dict:
    q = {"exp_id": exp.exp_id, "exp_session_id": sid, "type": DataManager.EXP_DATA}
    return exp.db_main.find_one(q)


def get_session_local(exp, sid) -> dict:
    path = exp.config.get("local_saving_agent", "path")
    path = exp.subpath(path)
    data = DataManager.iterate_local_data(DataManager.EXP_DATA, path)
    return next(s for s in data if s["exp_session_id"] == sid)


def get_data_of_session(exp, sid):
    if saving_method(exp) == "mongo":
        return get_session_mongo(exp, sid)
    elif saving_method(exp) == "local":
        return get_session_local(exp, sid)
